#include "QSearch.h"
#include "QSearchDlg.h"
#include "DialogSwitcher.h"


#define  WRONG_DWORD_VALUE          0xFEDC
#define  WRONG_COLOR_VALUE          0xFFFFFFFFL
#define  DEFAULT_COLOR_NOTFOUND     RGB(0xFF, 0xE0, 0xE0)
#define  DEFAULT_COLOR_EOF          RGB(0xE4, 0xFF, 0xE4)
#define  DEFAULT_COLOR_HIGHLIGHT    RGB(0xC0, 0xFF, 0xC0)
#define  DEFAULT_HIGHLIGHT_MARK_ID  1001
#define  DEFAULT_ALT_MATCHCASE      0x43  // 'C' (Case)
#define  DEFAULT_ALT_WHOLEWORD      0x57  // 'W' (Whole Word)
#define  DEFAULT_ALT_HIGHLIGHTALL   0x48  // 'H' (Highlight)

/* >>>>>>>>>>>>>>>>>>>>>>>> plugin state >>>>>>>>>>>>>>>>>>>>>>>> */
    void initializePluginState(PluginState* pPlugin)
    {
        pPlugin->bInitialized = FALSE;
        pPlugin->hInstanceDLL = NULL;
        pPlugin->hMainWnd = NULL;
        pPlugin->bAkelPadOnFinish = FALSE;
        pPlugin->wLangSystem = 0; /*GetUserDefaultLangID();*/
        pPlugin->pMainProcData = NULL;
        //pPlugin->pEditProcData = NULL;
        pPlugin->pFrameProcData = NULL;
    }
/* <<<<<<<<<<<<<<<<<<<<<<<< plugin state <<<<<<<<<<<<<<<<<<<<<<<< */


/* >>>>>>>>>>>>>>>>>>>>>>>> qsearch options >>>>>>>>>>>>>>>>>>>>>>>> */
    void initializeOptions(QSearchOpt* pOptions)
    {
        int i;
        for ( i = 0; i < OPTF_COUNT; i++ )
        {
            pOptions->dwFlags[i] = WRONG_DWORD_VALUE;
        }
        pOptions->dockRect.left = 0;
        pOptions->dockRect.right = 0;
        pOptions->dockRect.top = 0;
        pOptions->dockRect.bottom = 0;
        pOptions->colorNotFound = WRONG_COLOR_VALUE;
        pOptions->colorEOF = WRONG_COLOR_VALUE;
        pOptions->colorHighlight = WRONG_COLOR_VALUE;
        pOptions->dwHighlightMarkID = WRONG_DWORD_VALUE;
        pOptions->dwHighlightState = WRONG_DWORD_VALUE;
        pOptions->dwUseAltHotkeys = WRONG_DWORD_VALUE;
        pOptions->dwAltMatchCase = WRONG_DWORD_VALUE;
        pOptions->dwAltWholeWord = WRONG_DWORD_VALUE;
        pOptions->dwAltHighlightAll = WRONG_DWORD_VALUE;
    }

    void copyOptions(QSearchOpt* pOptDst, const QSearchOpt* pOptSrc)
    {
        int i;
        for ( i = 0; i < OPTF_COUNT; i++ )
        {
            pOptDst->dwFlags[i] = pOptSrc->dwFlags[i];
        }
        pOptDst->dockRect.left = pOptSrc->dockRect.left;
        pOptDst->dockRect.right = pOptSrc->dockRect.right;
        pOptDst->dockRect.top = pOptSrc->dockRect.top;
        pOptDst->dockRect.bottom = pOptSrc->dockRect.bottom;
        pOptDst->colorNotFound = pOptSrc->colorNotFound;
        pOptDst->colorEOF = pOptSrc->colorEOF;
        pOptDst->colorHighlight = pOptSrc->colorHighlight;
        pOptDst->dwHighlightMarkID = pOptSrc->dwHighlightMarkID;
        pOptDst->dwHighlightState = pOptSrc->dwHighlightState;
        pOptDst->dwUseAltHotkeys = pOptSrc->dwUseAltHotkeys;
        pOptDst->dwAltMatchCase = pOptSrc->dwAltMatchCase;
        pOptDst->dwAltWholeWord = pOptSrc->dwAltWholeWord;
        pOptDst->dwAltHighlightAll = pOptSrc->dwAltHighlightAll;
    }

    BOOL equalOptions(const QSearchOpt* pOpt1, const QSearchOpt* pOpt2)
    {
        int i;
        for ( i = 0; i < OPTF_COUNT; i++ )
        {
            if ( pOpt1->dwFlags[i] != pOpt2->dwFlags[i] )
                return FALSE;
        }
        
        if ( pOpt1->dockRect.left != pOpt2->dockRect.left )
            return FALSE;

        if ( pOpt1->dockRect.right != pOpt2->dockRect.right )
            return FALSE;

        if ( pOpt1->dockRect.top != pOpt2->dockRect.top )
            return FALSE;

        if ( pOpt1->dockRect.bottom != pOpt2->dockRect.bottom )
            return FALSE;

        if ( pOpt1->colorNotFound != pOpt2->colorNotFound )
            return FALSE;

        if ( pOpt1->colorEOF != pOpt2->colorEOF )
            return FALSE;

        if ( pOpt1->colorHighlight != pOpt2->colorHighlight )
            return FALSE;

        if ( pOpt1->dwHighlightMarkID != pOpt2->dwHighlightMarkID )
            return FALSE;

        if ( pOpt1->dwHighlightState != pOpt2->dwHighlightState )
            return FALSE;

        if ( pOpt1->dwUseAltHotkeys != pOpt2->dwUseAltHotkeys )
            return FALSE;

        if ( pOpt1->dwAltMatchCase != pOpt2->dwAltMatchCase )
            return FALSE;

        if ( pOpt1->dwAltWholeWord != pOpt2->dwAltWholeWord )
            return FALSE;

        if ( pOpt1->dwAltHighlightAll != pOpt2->dwAltHighlightAll )
            return FALSE;

        return TRUE;
    }
/* <<<<<<<<<<<<<<<<<<<<<<<< qsearch options <<<<<<<<<<<<<<<<<<<<<<<< */


// const
const char*    CSZ_QSEARCH = "QSearch";
const wchar_t* CWSZ_QSEARCH = L"QSearch";

const char*    CSZ_OPTIONS[OPT_TOTALCOUNT] = {
  /*  0 */  "srch_onthefly_mode",
  /*  1 */  "srch_from_beginning",
  /*  2 */  "srch_use_specialchars",
  /*  3 */  "srch_pickup_selection",
  /*  4 */  "srch_selfind_pickup",
  /*  5 */  "srch_stop_eof",
  /*  6 */  "srch_wnd_dockedtop",
  /*  7 */  "edit_focus_selectall",
  /*  8 */  "catch_main_f3",
  /*  9 */  "catch_main_esc",
  /* 10 */  "hotkey_closes_wnd",
  /* 11 */  "dock_rect_disabled",
  /* 12 */  "dock_rect",
  /* 13 */  "color_notfound",
  /* 14 */  "color_eof",
  /* 15 */  "color_highlight",
  /* 16 */  "highlight_mark_id",
  /* 17 */  "highlight_state",
  /* 18 */  "use_alt_hotkeys",
  /* 19 */  "alt_match_case",
  /* 20 */  "alt_whole_word",
  /* 21 */  "alt_highlight_all"
};

const wchar_t* CWSZ_OPTIONS[OPT_TOTALCOUNT] = {
  /*  0 */  L"srch_onthefly_mode",
  /*  1 */  L"srch_from_beginning",
  /*  2 */  L"srch_use_specialchars",
  /*  3 */  L"srch_pickup_selection",
  /*  4 */  L"srch_selfind_pickup",
  /*  5 */  L"srch_stop_eof",
  /*  6 */  L"srch_wnd_dockedtop",
  /*  7 */  L"edit_focus_selectall",
  /*  8 */  L"catch_main_f3",
  /*  9 */  L"catch_main_esc",
  /* 10 */  L"hotkey_closes_wnd",
  /* 11 */  L"dock_rect_disabled",
  /* 12 */  L"dock_rect",
  /* 13 */  L"color_notfound",
  /* 14 */  L"color_eof",
  /* 15 */  L"color_highlight",
  /* 16 */  L"highlight_mark_id",
  /* 17 */  L"highlight_state",
  /* 18 */  L"use_alt_hotkeys",
  /* 19 */  L"alt_match_case",
  /* 20 */  L"alt_whole_word",
  /* 21 */  L"alt_highlight_all"
};


// vars
PluginState     g_Plugin;
QSearchDlgState g_QSearchDlg;
QSearchOpt      g_Options;
QSearchOpt      g_Options0;
wchar_t         g_szFunctionQSearchW[128] = { 0 };
BOOL            g_bHighlightPlugin = FALSE;


// funcs
void  ReadOptions(void);
void  SaveOptions(void);
void  readBinaryA(HANDLE, const char*, void*, DWORD);
void  readBinaryW(HANDLE, const wchar_t*, void*, DWORD);
DWORD readDwordA(HANDLE, const char*, DWORD);
DWORD readDwordW(HANDLE, const wchar_t*, DWORD);
void  writeBinaryA(HANDLE, const char*, const void*, DWORD);
void  writeBinaryW(HANDLE, const wchar_t*, const void*, DWORD);
void  writeDwordA(HANDLE, const char*, DWORD);
void  writeDwordW(HANDLE, const wchar_t*, DWORD);
LRESULT CALLBACK NewMainProc(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK NewFrameProc(HWND, UINT, WPARAM, LPARAM);


int doQSearch(PLUGINDATA* pd, BOOL bInternalCall)
{
    if ( !g_Plugin.bInitialized )
    {
        Initialize(pd);
    }
    
    if ( !bInternalCall )
    {
        if ( pd->bOldWindows && (0 == ((LPSTR) g_szFunctionQSearchW)[0]) )
        {
            lstrcpyA( (LPSTR) g_szFunctionQSearchW, (LPCSTR) pd->pFunction );
        }
        else if ( 0 == ((LPWSTR) g_szFunctionQSearchW)[0] )
        {
            lstrcpyW( (LPWSTR) g_szFunctionQSearchW, (LPCWSTR) pd->pFunction );
        }
    }

    if ( !g_QSearchDlg.hDlg )
    {
        if ( g_Plugin.bOldWindows )
        {
            g_QSearchDlg.hDlg = CreateDialogA( 
                g_Plugin.hInstanceDLL, 
                MAKEINTRESOURCEA(IDD_QSEARCH),
                g_Plugin.hMainWnd,
                qsearchDlgProc );
        }
        else
        {
            g_QSearchDlg.hDlg = CreateDialogW( 
                g_Plugin.hInstanceDLL, 
                MAKEINTRESOURCEW(IDD_QSEARCH),
                g_Plugin.hMainWnd,
                qsearchDlgProc );
        }

        if ( g_QSearchDlg.hDlg )
        {
            if ( g_QSearchDlg.pDockData )
            {
                WPARAM dockAction;
                
                //g_QSearchDlg.bOnDlgStart = TRUE;
                if ( (g_QSearchDlg.pDockData->rcSize.right == 0) ||
                     (g_QSearchDlg.pDockData->rcSize.bottom == 0) )
                {
                    if ( GetWindowRect(g_QSearchDlg.hDlg, &g_QSearchDlg.pDockData->rcSize) )
                    {
                        g_QSearchDlg.pDockData->rcSize.right -= g_QSearchDlg.pDockData->rcSize.left;
                        g_QSearchDlg.pDockData->rcSize.bottom -= g_QSearchDlg.pDockData->rcSize.top;
                        g_QSearchDlg.pDockData->rcSize.left = 0;
                        g_QSearchDlg.pDockData->rcSize.top = 0;
                    }
                }
                g_QSearchDlg.pDockData->hWnd = g_QSearchDlg.hDlg;
                dockAction = bInternalCall ? DK_SUBCLASS : (DK_SUBCLASS | DK_SHOW);
                SendMessage( g_Plugin.hMainWnd, AKD_DOCK, 
                  dockAction, (LPARAM) g_QSearchDlg.pDockData );
            }
            else
            {
                SendMessage(g_QSearchDlg.hDlg, QSM_QUIT, 0, 0);
            }
        }
    }
    else if ( !bInternalCall )
    {
        if ( g_Options.dwFlags[OPTF_HOTKEY_CLOSES_WND] )
        {
            if ( IsWindowVisible(g_QSearchDlg.hDlg) )
            {
                SendMessage( g_QSearchDlg.hDlg, QSM_SHOWHIDE, FALSE, 0 );
                
                // Stay in memory, and show as non-active
                return UD_NONUNLOAD_NONACTIVE;
            }
        }
        SendMessage( g_QSearchDlg.hDlg, QSM_SHOWHIDE, TRUE, 0 );
    }
    
    // Stay in memory, and show as active
    return UD_NONUNLOAD_ACTIVE;
}

// Identification
void __declspec(dllexport) DllAkelPadID(PLUGINVERSION *pv)
{
    pv->dwAkelDllVersion  = AKELDLL;
    pv->dwExeMinVersion3x = MAKE_IDENTIFIER(3, 7, 4, 0);
    pv->dwExeMinVersion4x = MAKE_IDENTIFIER(4, 4, 0, 0);
    pv->pPluginName = "QSearch";
}

// Plugin extern function
void __declspec(dllexport) QSearch(PLUGINDATA* pd)
{
    if ( pd->lpbAutoLoad )
    {
        *(pd->lpbAutoLoad) = TRUE;
        return;
    }
    
    pd->nUnload = doQSearch(pd, FALSE);
}

// Plugin extern function
void __declspec(dllexport) SelFindNext(PLUGINDATA* pd)
{
    if ( pd->lpbAutoLoad )
    {
        *(pd->lpbAutoLoad) = FALSE;
        return;
    }

    doQSearch(pd, TRUE);
    if ( g_QSearchDlg.hDlg )
    {
        SendMessage( g_QSearchDlg.hDlg, QSM_SELFIND, FALSE, 0 );
    }

    // Stay in memory, and show as non-active
    pd->nUnload = UD_NONUNLOAD_NONACTIVE;
}

// Plugin extern function
void __declspec(dllexport) SelFindPrev(PLUGINDATA* pd)
{
    if ( pd->lpbAutoLoad )
    {
        *(pd->lpbAutoLoad) = FALSE;
        return;
    }

    doQSearch(pd, TRUE);
    if ( g_QSearchDlg.hDlg )
    {
        SendMessage( g_QSearchDlg.hDlg, QSM_SELFIND, TRUE, 0 );
    }

    // Stay in memory, and show as non-active
    pd->nUnload = UD_NONUNLOAD_NONACTIVE;
}

// Plugin extern function
void __declspec(dllexport) DialogSwitcher(PLUGINDATA* pd)
{
    if ( pd->lpbAutoLoad )
    {
        *(pd->lpbAutoLoad) = TRUE;
        return;
    }
    
    if ( dlgswtchIsActive() )
        dlgswtchUninitialize();
    else
        dlgswtchInitialize(pd);

    if ( dlgswtchIsActive() ) // ensure it's active
        pd->nUnload = UD_NONUNLOAD_ACTIVE;
    else
        pd->nUnload = UD_NONUNLOAD_NONACTIVE;
}

// Entry point
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD dwReason, LPVOID lpReserved)
{
    switch ( dwReason )
    {
        case DLL_PROCESS_ATTACH:
            g_szFunctionQSearchW[0] = 0;
            initializePluginState( &g_Plugin );
            initializeQSearchDlgState( &g_QSearchDlg );
            initializeOptions( &g_Options );
            initializeOptions( &g_Options0 );
            break;
        case DLL_PROCESS_DETACH:
            dlgswtchUninitialize();
            Uninitialize();
            break;
        case DLL_THREAD_ATTACH:
            break;
        case DLL_THREAD_DETACH:
            break;
        default:
            break;
    }
    return TRUE;
}

static LANGID getAkelPadLang(PLUGINDATA* pd)
{
    HANDLE hOptions = (HANDLE) SendMessage( 
      pd->hMainWnd, AKD_BEGINOPTIONS, POB_READ, (LPARAM) NULL );

    if ( hOptions ) 
    {
        wchar_t wszModule[MAX_PATH];

        wszModule[0] = 0;
        if ( pd->bOldWindows )
        {
            PLUGINOPTIONA poA;

            poA.pOptionName = "LanguageModule";
            poA.dwType = PO_STRING;
            poA.lpData = (LPBYTE) wszModule;
            poA.dwData = MAX_PATH;
            SendMessage( pd->hMainWnd,
              AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poA );
        }
        else
        {
            PLUGINOPTIONW poW;

            poW.pOptionName = L"LanguageModule";
            poW.dwType = PO_STRING;
            poW.lpData = (LPBYTE) wszModule;
            poW.dwData = MAX_PATH * sizeof(wchar_t);
            SendMessage( pd->hMainWnd,
              AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poW );
        }
        SendMessage(pd->hMainWnd, AKD_ENDOPTIONS, (WPARAM) hOptions, 0);
        if ( pd->bOldWindows )
        {
            if ( ((LPSTR) wszModule)[0] )
            {
                ((LPSTR) wszModule)[3] = 0;
                CharUpperA( (LPSTR) wszModule );
                if ( lstrcmpA( (LPCSTR) wszModule, "ENG" ) == 0 )
                    return LANG_ENGLISH;
                if ( lstrcmpA( (LPCSTR) wszModule, "RUS" ) == 0 )
                    return LANG_RUSSIAN;
                if ( lstrcmpA( (LPCSTR) wszModule, "UKR" ) == 0 )
                    return LANG_UKRAINIAN;
            }
        }
        else
        {
            if ( wszModule[0] )
            {
                wszModule[3] = 0;
                CharUpperW( wszModule );
                if ( lstrcmpW( wszModule, L"ENG" ) == 0 )
                    return LANG_ENGLISH;
                if ( lstrcmpW( wszModule, L"RUS" ) == 0 )
                    return LANG_RUSSIAN;
                if ( lstrcmpW( wszModule, L"UKR" ) == 0 )
                    return LANG_UKRAINIAN;
            }
        }
    }
    
    return pd->wLangSystem;
}

static BOOL findHighlightPlugin(PLUGINDATA* pd)
{
    BOOL bExists = FALSE;
    
    if ( pd->bOldWindows )
    {
        char             szPluginPath[MAX_PATH + 1];
        HANDLE           hFind;
        WIN32_FIND_DATAA findData;

        lstrcpyA( szPluginPath, (const char *) pd->pAkelDir );
        lstrcatA( szPluginPath, "\\AkelFiles\\Plugs\\HighLight.dll" );
        
        hFind = FindFirstFileA(szPluginPath, &findData);
        if ( hFind && (hFind != INVALID_HANDLE_VALUE) )
        {
            if ( !(findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) )
            {
                bExists = TRUE;
            }
            FindClose(hFind);
        }
    }
    else
    {
        wchar_t          szPluginPath[MAX_PATH + 1];
        HANDLE           hFind;
        WIN32_FIND_DATAW findData;

        lstrcpyW( szPluginPath, (const wchar_t *) pd->pAkelDir );
        lstrcatW( szPluginPath, L"\\AkelFiles\\Plugs\\HighLight.dll" );

        hFind = FindFirstFileW(szPluginPath, &findData);
        if ( hFind && (hFind != INVALID_HANDLE_VALUE) )
        {
            if ( !(findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) )
            {
                bExists = TRUE;
            }
            FindClose(hFind);
        }
    }

    return bExists;
}

void Initialize(PLUGINDATA* pd)
{
    if ( !g_Plugin.bInitialized )
    {
        DOCK dk = { 0 };

        g_Plugin.bOldWindows  = pd->bOldWindows;
        g_Plugin.bOldRichEdit = pd->bOldRichEdit;
        g_Plugin.bMDI         = pd->bMDI;
        g_Plugin.hInstanceDLL = pd->hInstanceDLL;
        g_Plugin.hMainWnd     = pd->hMainWnd;
        g_Plugin.wLangSystem  = getAkelPadLang(pd);

        g_bHighlightPlugin = findHighlightPlugin(pd);

        ReadOptions();
        
        dk.nSide = g_Options.dwFlags[OPTF_SRCH_WND_DOCKEDTOP] ? DKS_TOP : DKS_BOTTOM;
        dk.dwFlags = DKF_DRAGDROP | DKF_NODROPLEFT | DKF_NODROPRIGHT | DKF_FIXEDSIZE | DKF_HIDDEN;
        if ( !g_Options.dwFlags[OPTF_DOCK_RECT_DISABLED] )
        {
            dk.rcSize.left = g_Options.dockRect.left;
            dk.rcSize.right = g_Options.dockRect.right;
            dk.rcSize.top = g_Options.dockRect.top;
            dk.rcSize.bottom = g_Options.dockRect.bottom;
        }
        g_QSearchDlg.pDockData = (DOCK *) SendMessage( 
          g_Plugin.hMainWnd, AKD_DOCK, DK_ADD, (LPARAM) &dk );
        
        // SubClass main window
        g_Plugin.pMainProcData = NULL;
        SendMessage( g_Plugin.hMainWnd, AKD_SETMAINPROC,
          (WPARAM) NewMainProc, (LPARAM) &g_Plugin.pMainProcData );

        if ( g_Plugin.bMDI )
        {
            // SubClass frame window
            g_Plugin.pFrameProcData = NULL;
            SendMessage( g_Plugin.hMainWnd, AKD_SETFRAMEPROC,
              (WPARAM) NewFrameProc, (LPARAM) &g_Plugin.pFrameProcData );
        }
    }
    g_Plugin.bInitialized = TRUE;
}

void Uninitialize(void)
{
    if ( g_Plugin.bInitialized )
    {
        SaveOptions();
        
        if ( g_Plugin.pMainProcData )
        {
            // Remove subclass (main window)
            SendMessage( g_Plugin.hMainWnd, AKD_SETMAINPROC, 
              (WPARAM) 0, (LPARAM) &g_Plugin.pMainProcData );
            g_Plugin.pMainProcData = NULL;
        }

        if ( g_Plugin.bMDI && g_Plugin.pFrameProcData )
        {
            // Remove subclass (frame window)
            SendMessage( g_Plugin.hMainWnd, AKD_SETFRAMEPROC,
              (WPARAM) 0, (LPARAM) &g_Plugin.pFrameProcData );
            g_Plugin.pFrameProcData = NULL;
        }

        if ( g_QSearchDlg.hDlg )
        {
            SendMessage(g_QSearchDlg.hDlg, QSM_QUIT, 0, 0);
        }
        
        if ( g_QSearchDlg.pDockData )
        {
            SendMessage( g_Plugin.hMainWnd, AKD_DOCK, 
              DK_DELETE, (LPARAM) g_QSearchDlg.pDockData );
        }
        
        g_Plugin.bInitialized = FALSE;
    }
}

LRESULT CALLBACK NewMainProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch ( uMsg )
    {
        case WM_COMMAND:
        {
            switch ( LOWORD(wParam) )
            {
                case IDM_FILE_EXIT:
                    if ( HIWORD(wParam) == 1 )
                    {
                        // this code works when Esc is pressed in AkelPad
                        if ( g_Options.dwFlags[OPTF_CATCH_MAIN_ESC] && 
                             g_QSearchDlg.hDlg && 
                             IsWindowVisible(g_QSearchDlg.hDlg) )
                        {
                            SendMessage( g_QSearchDlg.hDlg, QSM_SHOWHIDE, FALSE, 0 );
                            return 0;
                        }
                    }
                    break;
                case IDM_EDIT_FINDNEXTDOWN:
                case IDM_EDIT_FINDNEXTUP:
                    if ( g_Options.dwFlags[OPTF_CATCH_MAIN_F3] )
                    {
                        if ( g_QSearchDlg.bQSearching )
                        {
                            SendMessage( g_QSearchDlg.hDlg, QSM_FINDNEXT, 0, 0 );
                            return 0;
                        }
                    }
                    break;
                case IDM_EDIT_FIND:
                case IDM_EDIT_REPLACE:
                    // AkelPad's Find or Replace Dialog is shown
                    g_QSearchDlg.bQSearching = FALSE;
                    break;
                default:
                    break;
            }
            break;
        }
        case AKDN_SIZE:
        {
            if ( g_QSearchDlg.pDockData )
            {
                if ( g_QSearchDlg.pDockData->nSide == DKS_TOP )
                    g_Options.dwFlags[OPTF_SRCH_WND_DOCKEDTOP] = 1;
                else
                    g_Options.dwFlags[OPTF_SRCH_WND_DOCKEDTOP] = 0;
            }
            break;
        }
        case AKDN_MAIN_ONFINISH:
        {
            LRESULT lResult = 0;

            if ( g_Plugin.pMainProcData && g_Plugin.pMainProcData->NextProc )
                lResult = g_Plugin.pMainProcData->NextProc(hWnd, uMsg, wParam, lParam);
            
            g_Plugin.bAkelPadOnFinish = TRUE;
            dlgswtchUninitialize();
            if ( g_QSearchDlg.hDlg )
            {
                SendMessage( g_QSearchDlg.hDlg, QSM_QUIT, 0, 0 );
            }
            
            return lResult;
        }
        default:
            break;
    }
    
    if ( g_Plugin.pMainProcData && g_Plugin.pMainProcData->NextProc )
        return g_Plugin.pMainProcData->NextProc(hWnd, uMsg, wParam, lParam);
    else
        return 0;
}

LRESULT CALLBACK NewFrameProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch ( uMsg )
    {
        case WM_MDIACTIVATE:
            if ( g_QSearchDlg.hDlg )
            {
                SendMessage( g_QSearchDlg.hDlg, QSM_SETNOTFOUND, FALSE, 0 );
            }
            break;
        default:
            break;
    }

    if ( g_Plugin.pFrameProcData && g_Plugin.pFrameProcData->NextProc )
        return g_Plugin.pFrameProcData->NextProc(hWnd, uMsg, wParam, lParam);
    else
        return 0;
}


void ReadOptions(void)
{
    HANDLE hOptions;

    if ( g_Plugin.bOldWindows )
    {
        if ( hOptions = (HANDLE) SendMessage(g_Plugin.hMainWnd,
               AKD_BEGINOPTIONS, POB_READ, (LPARAM) CSZ_QSEARCH) )
        {
            int i;
            for ( i = 0; i < OPTF_COUNT; i++ )
            {
                g_Options.dwFlags[i] = readDwordA( hOptions, 
                  CSZ_OPTIONS[i], WRONG_DWORD_VALUE );
            }

            if ( !g_Options.dwFlags[OPTF_DOCK_RECT_DISABLED] )
            {
                readBinaryA( hOptions, CSZ_OPTIONS[OPT_DOCK_RECT],
                  &g_Options.dockRect, sizeof(RECT) );
            }

            readBinaryA( hOptions, CSZ_OPTIONS[OPT_COLOR_NOTFOUND],
              &g_Options.colorNotFound, sizeof(COLORREF) );

            readBinaryA( hOptions, CSZ_OPTIONS[OPT_COLOR_EOF],
              &g_Options.colorEOF, sizeof(COLORREF) );

            readBinaryA( hOptions, CSZ_OPTIONS[OPT_COLOR_HIGHLIGHT],
              &g_Options.colorHighlight, sizeof(COLORREF) );

            g_Options.dwHighlightMarkID = readDwordA( hOptions,
              CSZ_OPTIONS[OPT_HIGHLIGHT_MARK_ID], WRONG_DWORD_VALUE );

            g_Options.dwHighlightState = readDwordA( hOptions,
              CSZ_OPTIONS[OPT_HIGHLIGHT_STATE], WRONG_DWORD_VALUE );

            g_Options.dwUseAltHotkeys = readDwordA( hOptions, 
              CSZ_OPTIONS[OPT_USE_ALT_HOTKEYS], WRONG_DWORD_VALUE );

            g_Options.dwAltMatchCase = readDwordA( hOptions, 
              CSZ_OPTIONS[OPT_ALT_MATCHCASE], WRONG_DWORD_VALUE );

            g_Options.dwAltWholeWord = readDwordA( hOptions, 
              CSZ_OPTIONS[OPT_ALT_WHOLEWORD], WRONG_DWORD_VALUE );

            g_Options.dwAltHighlightAll = readDwordA( hOptions, 
              CSZ_OPTIONS[OPT_ALT_HIGHLIGHTALL], WRONG_DWORD_VALUE );
            
            // all options have been read
            SendMessage(g_Plugin.hMainWnd, AKD_ENDOPTIONS, (WPARAM) hOptions, 0);
        }
    }
    else
    {
        if ( hOptions = (HANDLE) SendMessage(g_Plugin.hMainWnd,
               AKD_BEGINOPTIONS, POB_READ, (LPARAM) CWSZ_QSEARCH) )
        {
            int i;
            for ( i = 0; i < OPTF_COUNT; i++ )
            {
                g_Options.dwFlags[i] = readDwordW( hOptions, 
                  CWSZ_OPTIONS[i], WRONG_DWORD_VALUE );
            }

            if ( !g_Options.dwFlags[OPTF_DOCK_RECT_DISABLED] )
            {
                readBinaryW( hOptions, CWSZ_OPTIONS[OPT_DOCK_RECT],
                  &g_Options.dockRect, sizeof(RECT) );
            }
            
            readBinaryW( hOptions, CWSZ_OPTIONS[OPT_COLOR_NOTFOUND],
              &g_Options.colorNotFound, sizeof(COLORREF) );

            readBinaryW( hOptions, CWSZ_OPTIONS[OPT_COLOR_EOF],
              &g_Options.colorEOF, sizeof(COLORREF) );
            
            readBinaryW( hOptions, CWSZ_OPTIONS[OPT_COLOR_HIGHLIGHT],
              &g_Options.colorHighlight, sizeof(COLORREF) );

            g_Options.dwHighlightMarkID = readDwordW( hOptions,
              CWSZ_OPTIONS[OPT_HIGHLIGHT_MARK_ID], WRONG_DWORD_VALUE );

            g_Options.dwHighlightState = readDwordW( hOptions,
              CWSZ_OPTIONS[OPT_HIGHLIGHT_STATE], WRONG_DWORD_VALUE );

            g_Options.dwUseAltHotkeys = readDwordW( hOptions, 
              CWSZ_OPTIONS[OPT_USE_ALT_HOTKEYS], WRONG_DWORD_VALUE );

            g_Options.dwAltMatchCase = readDwordW( hOptions, 
              CWSZ_OPTIONS[OPT_ALT_MATCHCASE], WRONG_DWORD_VALUE );

            g_Options.dwAltWholeWord = readDwordW( hOptions, 
              CWSZ_OPTIONS[OPT_ALT_WHOLEWORD], WRONG_DWORD_VALUE );

            g_Options.dwAltHighlightAll = readDwordW( hOptions, 
              CWSZ_OPTIONS[OPT_ALT_HIGHLIGHTALL], WRONG_DWORD_VALUE );

            // all options have been read
            SendMessage(g_Plugin.hMainWnd, AKD_ENDOPTIONS, (WPARAM) hOptions, 0);
        }
    }
    copyOptions( &g_Options0, &g_Options );
    
    // normalize option values
    if ( g_Options.dwFlags[OPTF_SRCH_ONTHEFLY_MODE] == WRONG_DWORD_VALUE )
        g_Options.dwFlags[OPTF_SRCH_ONTHEFLY_MODE] = 1;

    /*if ( g_Options.dwFlags[OPTF_SRCH_ONTHEFLY_MODE] )
        g_Options.dwFlags[OPTF_SRCH_USE_SPECIALCHARS] = 0;*/

    if ( g_Options.dwFlags[OPTF_SRCH_FROM_BEGINNING] == WRONG_DWORD_VALUE )
        g_Options.dwFlags[OPTF_SRCH_FROM_BEGINNING] = 0;

    if ( g_Options.dwFlags[OPTF_SRCH_USE_SPECIALCHARS] == WRONG_DWORD_VALUE )
        g_Options.dwFlags[OPTF_SRCH_USE_SPECIALCHARS] = 0;
    
    if ( g_Options.dwFlags[OPTF_SRCH_PICKUP_SELECTION] == WRONG_DWORD_VALUE )
        g_Options.dwFlags[OPTF_SRCH_PICKUP_SELECTION] = 0;

    if ( g_Options.dwFlags[OPTF_SRCH_SELFIND_PICKUP] == WRONG_DWORD_VALUE )
        g_Options.dwFlags[OPTF_SRCH_SELFIND_PICKUP] = 0;
    
    if ( g_Options.dwFlags[OPTF_SRCH_STOP_EOF] == WRONG_DWORD_VALUE )
        g_Options.dwFlags[OPTF_SRCH_STOP_EOF] = 0;

    if ( g_Options.dwFlags[OPTF_SRCH_WND_DOCKEDTOP] == WRONG_DWORD_VALUE )
        g_Options.dwFlags[OPTF_SRCH_WND_DOCKEDTOP] = 0;

    if ( g_Options.dwFlags[OPTF_EDIT_FOCUS_SELECTALL] == WRONG_DWORD_VALUE )
        g_Options.dwFlags[OPTF_EDIT_FOCUS_SELECTALL] = 0;

    if ( g_Options.dwFlags[OPTF_CATCH_MAIN_F3] == WRONG_DWORD_VALUE )
        g_Options.dwFlags[OPTF_CATCH_MAIN_F3] = 0;

    if ( g_Options.dwFlags[OPTF_CATCH_MAIN_ESC] == WRONG_DWORD_VALUE )
        g_Options.dwFlags[OPTF_CATCH_MAIN_ESC] = 0;

    if ( g_Options.dwFlags[OPTF_HOTKEY_CLOSES_WND] == WRONG_DWORD_VALUE )
        g_Options.dwFlags[OPTF_HOTKEY_CLOSES_WND] = 0;
    
    if ( g_Options.dwFlags[OPTF_DOCK_RECT_DISABLED] == WRONG_DWORD_VALUE )
        g_Options.dwFlags[OPTF_DOCK_RECT_DISABLED] = 0;

    if ( g_Options.colorNotFound == WRONG_COLOR_VALUE )
        g_Options.colorNotFound = DEFAULT_COLOR_NOTFOUND;

    if ( g_Options.colorEOF == WRONG_COLOR_VALUE )
        g_Options.colorEOF = DEFAULT_COLOR_EOF;

    if ( g_Options.colorHighlight == WRONG_COLOR_VALUE )
        g_Options.colorHighlight = DEFAULT_COLOR_HIGHLIGHT;

    if ( g_Options.dwHighlightMarkID == WRONG_DWORD_VALUE )
        g_Options.dwHighlightMarkID = DEFAULT_HIGHLIGHT_MARK_ID;

    if ( g_Options.dwHighlightState == WRONG_DWORD_VALUE )
        g_Options.dwHighlightState = 0;

    if ( g_Options.dwUseAltHotkeys == WRONG_DWORD_VALUE )
        g_Options.dwUseAltHotkeys = 0;

    if ( g_Options.dwAltMatchCase == WRONG_DWORD_VALUE )
        g_Options.dwAltMatchCase = DEFAULT_ALT_MATCHCASE;

    if ( g_Options.dwAltWholeWord == WRONG_DWORD_VALUE )
        g_Options.dwAltWholeWord = DEFAULT_ALT_WHOLEWORD;

    if ( g_Options.dwAltHighlightAll == WRONG_DWORD_VALUE )
        g_Options.dwAltHighlightAll = DEFAULT_ALT_HIGHLIGHTALL;
}

void SaveOptions(void)
{
    if ( g_QSearchDlg.pDockData )
    {
        if ( g_QSearchDlg.pDockData->nSide == DKS_TOP )
            g_Options.dwFlags[OPTF_SRCH_WND_DOCKEDTOP] = 1;
        else
            g_Options.dwFlags[OPTF_SRCH_WND_DOCKEDTOP] = 0;

        if ( !g_Options.dwFlags[OPTF_DOCK_RECT_DISABLED] )
        {
            g_Options.dockRect.left = g_QSearchDlg.pDockData->rcSize.left;
            g_Options.dockRect.right = g_QSearchDlg.pDockData->rcSize.right;
            g_Options.dockRect.top = g_QSearchDlg.pDockData->rcSize.top;
            g_Options.dockRect.bottom = g_QSearchDlg.pDockData->rcSize.bottom;
        }
    }
    
    if ( !equalOptions(&g_Options0, &g_Options) )
    {
        HANDLE hOptions;

        if ( g_Plugin.bOldWindows )
        {
            if ( hOptions = (HANDLE) SendMessage(g_Plugin.hMainWnd, 
                   AKD_BEGINOPTIONS, POB_SAVE, (LPARAM) CSZ_QSEARCH) )
            {
                int i;
                for ( i = 0; i < OPTF_COUNT; i++ )
                {
                    writeDwordA( hOptions, 
                      CSZ_OPTIONS[i], g_Options.dwFlags[i] );
                }

                if ( !g_Options.dwFlags[OPTF_DOCK_RECT_DISABLED] )
                {
                    writeBinaryA( hOptions, CSZ_OPTIONS[OPT_DOCK_RECT],
                      &g_Options.dockRect, sizeof(RECT) );
                }

                writeBinaryA( hOptions, CSZ_OPTIONS[OPT_COLOR_NOTFOUND],
                  &g_Options.colorNotFound, sizeof(COLORREF) );

                writeBinaryA( hOptions, CSZ_OPTIONS[OPT_COLOR_EOF],
                  &g_Options.colorEOF, sizeof(COLORREF) );

                writeBinaryA( hOptions, CSZ_OPTIONS[OPT_COLOR_HIGHLIGHT],
                  &g_Options.colorHighlight, sizeof(COLORREF) );

                writeDwordA( hOptions, CSZ_OPTIONS[OPT_HIGHLIGHT_MARK_ID],
                  g_Options.dwHighlightMarkID );

                if ( (g_Options.dwHighlightState & HLS_SET_ALWAYS) != HLS_SET_ALWAYS )
                {    
                    writeDwordA( hOptions, CSZ_OPTIONS[OPT_HIGHLIGHT_STATE],
                      g_Options.dwHighlightState );
                }

                writeDwordA( hOptions, CSZ_OPTIONS[OPT_USE_ALT_HOTKEYS],
                  g_Options.dwUseAltHotkeys );

                writeDwordA( hOptions, CSZ_OPTIONS[OPT_ALT_MATCHCASE], 
                  g_Options.dwAltMatchCase );

                writeDwordA( hOptions, CSZ_OPTIONS[OPT_ALT_WHOLEWORD], 
                  g_Options.dwAltWholeWord );

                writeDwordA( hOptions, CSZ_OPTIONS[OPT_ALT_HIGHLIGHTALL], 
                  g_Options.dwAltHighlightAll );

                // all options have been saved
                SendMessage(g_Plugin.hMainWnd, AKD_ENDOPTIONS, (WPARAM) hOptions, 0);
            }
        }
        else
        {
            if ( hOptions = (HANDLE) SendMessage(g_Plugin.hMainWnd, 
                   AKD_BEGINOPTIONS, POB_SAVE, (LPARAM) CWSZ_QSEARCH) )
            {
                int i;
                for ( i = 0; i < OPTF_COUNT; i++ )
                {
                    writeDwordW( hOptions, 
                      CWSZ_OPTIONS[i], g_Options.dwFlags[i] );
                }

                if ( !g_Options.dwFlags[OPTF_DOCK_RECT_DISABLED] )
                {
                    writeBinaryW( hOptions, CWSZ_OPTIONS[OPT_DOCK_RECT],
                      &g_Options.dockRect, sizeof(RECT) );
                }

                writeBinaryW( hOptions, CWSZ_OPTIONS[OPT_COLOR_NOTFOUND],
                  &g_Options.colorNotFound, sizeof(COLORREF) );

                writeBinaryW( hOptions, CWSZ_OPTIONS[OPT_COLOR_EOF],
                  &g_Options.colorEOF, sizeof(COLORREF) );

                writeBinaryW( hOptions, CWSZ_OPTIONS[OPT_COLOR_HIGHLIGHT],
                  &g_Options.colorHighlight, sizeof(COLORREF) );

                writeDwordW( hOptions, CWSZ_OPTIONS[OPT_HIGHLIGHT_MARK_ID],
                  g_Options.dwHighlightMarkID );

                if ( (g_Options.dwHighlightState & HLS_SET_ALWAYS) != HLS_SET_ALWAYS )
                {    
                    writeDwordW( hOptions, CWSZ_OPTIONS[OPT_HIGHLIGHT_STATE],
                      g_Options.dwHighlightState );
                }

                writeDwordW( hOptions, CWSZ_OPTIONS[OPT_USE_ALT_HOTKEYS],
                  g_Options.dwUseAltHotkeys );

                writeDwordW( hOptions, CWSZ_OPTIONS[OPT_ALT_MATCHCASE], 
                  g_Options.dwAltMatchCase );

                writeDwordW( hOptions, CWSZ_OPTIONS[OPT_ALT_WHOLEWORD], 
                  g_Options.dwAltWholeWord );

                writeDwordW( hOptions, CWSZ_OPTIONS[OPT_ALT_HIGHLIGHTALL], 
                  g_Options.dwAltHighlightAll );

                // all options have been saved
                SendMessage(g_Plugin.hMainWnd, AKD_ENDOPTIONS, (WPARAM) hOptions, 0);
            }
        }
        copyOptions( &g_Options0, &g_Options );
    }
}

void readBinaryA(HANDLE hOptions, const char* szOptionNameA, 
                 void* pData, DWORD dwDataSize)
{
    PLUGINOPTIONA poA;
    
    poA.pOptionName = (char *) szOptionNameA;
    poA.dwType = PO_BINARY;
    poA.lpData = (BYTE *) pData;
    poA.dwData = dwDataSize;
    SendMessage( g_Plugin.hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poA );
}

void readBinaryW(HANDLE hOptions, const wchar_t* szOptionNameW, 
                 void* pData, DWORD dwDataSize)
{
    PLUGINOPTIONW poW;
    
    poW.pOptionName = (wchar_t *) szOptionNameW;
    poW.dwType = PO_BINARY;
    poW.lpData = (BYTE *) pData;
    poW.dwData = dwDataSize;
    SendMessage( g_Plugin.hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poW );
}

DWORD readDwordA(HANDLE hOptions, const char* szOptionNameA, DWORD dwDefaultVal)
{
    PLUGINOPTIONA poA;
    
    poA.pOptionName = (char *) szOptionNameA;
    poA.dwType = PO_DWORD;
    poA.lpData = (BYTE *) &dwDefaultVal;
    poA.dwData = sizeof(DWORD);
    SendMessage( g_Plugin.hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poA );
    return dwDefaultVal;
}

DWORD readDwordW(HANDLE hOptions, const wchar_t* szOptionNameW, DWORD dwDefaultVal)
{
    PLUGINOPTIONW poW;

    poW.pOptionName = (wchar_t *) szOptionNameW;
    poW.dwType = PO_DWORD;
    poW.lpData = (BYTE *) &dwDefaultVal;
    poW.dwData = sizeof(DWORD);
    SendMessage( g_Plugin.hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poW );
    return dwDefaultVal;
}

void  writeBinaryA(HANDLE hOptions, const char* szOptionNameA, 
                   const void* pData, DWORD dwDataSize)
{
    PLUGINOPTIONA poA;

    poA.pOptionName = (char *) szOptionNameA;
    poA.dwType = PO_BINARY;
    poA.lpData = (BYTE *) pData;
    poA.dwData = dwDataSize;
    SendMessage( g_Plugin.hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poA );
}

void  writeBinaryW(HANDLE hOptions, const wchar_t* szOptionNameW, 
                   const void* pData, DWORD dwDataSize)
{
    PLUGINOPTIONW poW;

    poW.pOptionName = (wchar_t *) szOptionNameW;
    poW.dwType = PO_BINARY;
    poW.lpData = (BYTE *) pData;
    poW.dwData = dwDataSize;
    SendMessage( g_Plugin.hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poW );
}

void writeDwordA(HANDLE hOptions, const char* szOptionNameA, DWORD dwVal)
{
    PLUGINOPTIONA poA;

    poA.pOptionName = (char *) szOptionNameA;
    poA.dwType = PO_DWORD;
    poA.lpData = (BYTE *) &dwVal;
    poA.dwData = sizeof(DWORD);
    SendMessage( g_Plugin.hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poA );
}

void writeDwordW(HANDLE hOptions, const wchar_t* szOptionNameW, DWORD dwVal)
{
    PLUGINOPTIONW poW;

    poW.pOptionName = (wchar_t *) szOptionNameW;
    poW.dwType = PO_DWORD;
    poW.lpData = (BYTE *) &dwVal;
    poW.dwData = sizeof(DWORD);
    SendMessage( g_Plugin.hMainWnd, AKD_OPTION, (WPARAM) hOptions, (LPARAM) &poW );
}

