#include "QSearch.h"
#include "QSearchDlg.h"
#include "QSearchLng.h"


#define  QSEARCH_FIRST       0x0001
#define  QSEARCH_NEXT        0x0002
#define  QSEARCH_SEL         0x0100
#define  QSEARCH_SEL_FINDUP  0x0200
#define  QSEARCH_NOFINDUP    0x1000
#define  QSEARCH_NOFINDBEGIN 0x2000

#define  QSEARCH_EOF_DOWN    0x0001
#define  QSEARCH_EOF_UP      0x0002

#define  VK_QS_FINDBEGIN     VK_CONTROL  // Ctrl
#define  VK_QS_FINDUP        VK_SHIFT    // Shift
#define  VK_QS_PICKUPTEXT    VK_MENU     // Alt


/* >>>>>>>>>>>>>>>>>>>>>>>> highlight plugin >>>>>>>>>>>>>>>>>>>>>>>> */    
typedef struct sDLLECHIGHLIGHT_MARK {
    DWORD dwStructSize;
    int nAction;
    unsigned char *pColorText;
    unsigned char *pColorBk;
    BOOL bMatchCase;
    DWORD dwFontStyle;
    DWORD dwMarkID;
} DLLECHIGHLIGHT_MARK;

typedef struct sDLLECHIGHLIGHT_UNMARK {
    DWORD dwStructSize;
    int nAction;
    DWORD dwMarkID;
} DLLECHIGHLIGHT_UNMARK;

#define DLLA_HIGHLIGHT_SETEXTENTION     1
#define DLLA_HIGHLIGHT_MARKSELECTION    2
#define DLLA_HIGHLIGHT_UNMARKSELECTION  3
#define DLLA_HIGHLIGHT_CLEARCACHE       4

const char*    cszHighlightMainA = "HighLight::Main";
const wchar_t* cszHighlightMainW = L"HighLight::Main";
/* <<<<<<<<<<<<<<<<<<<<<<<< highlight plugin <<<<<<<<<<<<<<<<<<<<<<<< */


/* >>>>>>>>>>>>>>>>>>>>>>>> qsearchdlg state >>>>>>>>>>>>>>>>>>>>>>>> */
    void initializeQSearchDlgState(QSearchDlgState* pQSearchDlg)
    {
        pQSearchDlg->hDlg = NULL;
        pQSearchDlg->bMatchCase = FALSE;
        pQSearchDlg->szFindTextW[0] = 0;
        //pQSearchDlg->bOnDlgStart = FALSE;
        pQSearchDlg->bQSearching = FALSE;
        pQSearchDlg->pDockData = NULL;
    }
/* <<<<<<<<<<<<<<<<<<<<<<<< qsearchdlg state <<<<<<<<<<<<<<<<<<<<<<<< */


// extern vars
extern PluginState     g_Plugin;
extern QSearchDlgState g_QSearchDlg;
extern QSearchOpt      g_Options;
extern wchar_t         g_szFunctionQSearchW[128];
extern BOOL            g_bHighlightPlugin;


// static (local) vars
static WNDPROC prev_editWndProc = NULL;
static WNDPROC prev_btnWndProc = NULL;
static BOOL    qs_bEditIsActive = TRUE;
static BOOL    qs_bEditCanBeNonActive = TRUE;
static BOOL    qs_bEditSelJustChanged = FALSE;
static BOOL    qs_bEditNotFound = FALSE;
static BOOL    qs_bEditIsEOF = FALSE;
static int     qs_nEditEOF = 0;
static BOOL    qs_bEditTextChanged = TRUE;
static BOOL    qs_bEditTrackMouse = FALSE;
static BOOL    qs_bBtnTrackMouse = FALSE;
static BOOL    qs_bHotKeyPressedOnShow = FALSE;
static BOOL    qs_bMatchCaseJustChanged = FALSE;
static DWORD   qs_dwHotKey = 0;


// funcs
HWND qsearchDoInitToolTip(HWND hDlg, HWND hEdit);
void qsearchDoQuit(HWND hEdit, HWND hToolTip, HBRUSH hBrush1, HBRUSH hBrush2);
void qsearchDoSearchText(HWND hEdit, DWORD dwParams);
void qsearchDoSelFind(HWND hEdit, BOOL bFindPrev);
void qsearchDoSetNotFound(HWND hEdit, BOOL bNotFound, BOOL bEOF);
void qsearchDoShowHide(HWND hDlg, BOOL bShow);
void qsearchDoTryHighlightAll(HWND hDlg);
void qsearchDoTryUnhighlightAll();

static void qsdlgShowHideWholeWordCheckBox(HWND hDlg)
{
    static int nChHighlightAllLeft0 = 0;
    HWND hChWholeWord;
    HWND hChHighlightAll;
    RECT rectDlg;
    RECT rectChWholeWord;
    RECT rectChHighlightAll;
    
    hChWholeWord = GetDlgItem(hDlg, IDC_CH_WHOLEWORD);
    hChHighlightAll = GetDlgItem(hDlg, IDC_CH_HIGHLIGHTALL);
    GetWindowRect(hDlg, &rectDlg);
    GetWindowRect(hChWholeWord, &rectChWholeWord);
    GetWindowRect(hChHighlightAll, &rectChHighlightAll);

    if ( nChHighlightAllLeft0 == 0 )
        nChHighlightAllLeft0 = rectChHighlightAll.left - rectDlg.left;

    ShowWindow( 
        hChWholeWord,
        g_Options.dwFlags[OPTF_SRCH_USE_SPECIALCHARS] ? SW_HIDE : SW_SHOWNORMAL
    );

    MoveWindow(
        hChHighlightAll,
        g_Options.dwFlags[OPTF_SRCH_USE_SPECIALCHARS] ? 
          (rectChWholeWord.left - rectDlg.left) : nChHighlightAllLeft0,
        rectChHighlightAll.top - rectDlg.top,
        rectChHighlightAll.right - rectChHighlightAll.left,
        rectChHighlightAll.bottom - rectChHighlightAll.top,
        TRUE
    );
}

static DWORD getQSearchHotKey(void)
{
    if ( g_szFunctionQSearchW[0] )
    {
        if ( g_Plugin.bOldWindows )
        {
            PLUGINFUNCTIONA* pfA;

            pfA = (PLUGINFUNCTIONA *) SendMessage( g_Plugin.hMainWnd, 
              AKD_DLLFIND, (WPARAM) g_szFunctionQSearchW, 0 );
            if ( pfA )
            {
                return pfA->wHotkey;
            }
        }
        else
        {
            PLUGINFUNCTIONW* pfW;

            pfW = (PLUGINFUNCTIONW *) SendMessage( g_Plugin.hMainWnd, 
              AKD_DLLFIND, (WPARAM) g_szFunctionQSearchW, 0 );
            if ( pfW )
            {
                return pfW->wHotkey;
            }
        }
    }
    return 0;
}

static BOOL isQSearchHotKeyPressed(void)
{
    if ( qs_dwHotKey )
    {
        if ( GetKeyState(LOBYTE(qs_dwHotKey)) & 0x80 )
        {
            if ( GetKeyState(VK_MENU) & 0x80 )
            {
                if ( (HIBYTE(qs_dwHotKey) & HOTKEYF_ALT) != HOTKEYF_ALT )
                    return FALSE;
            }
            else
            {
                if ( HIBYTE(qs_dwHotKey) & HOTKEYF_ALT )
                    return FALSE;
            }
                
            if ( GetKeyState(VK_CONTROL) & 0x80 )
            {
                if ( (HIBYTE(qs_dwHotKey) & HOTKEYF_CONTROL) != HOTKEYF_CONTROL )
                    return FALSE;
            }
            else
            {
                if ( HIBYTE(qs_dwHotKey) & HOTKEYF_CONTROL )
                    return FALSE;
            }
                
            if ( GetKeyState(VK_SHIFT) & 0x80 )
            {
                if ( (HIBYTE(qs_dwHotKey) & HOTKEYF_SHIFT) != HOTKEYF_SHIFT )
                    return FALSE;
            }
            else
            {
                if ( HIBYTE(qs_dwHotKey) & HOTKEYF_SHIFT )
                    return FALSE;
            }

            return TRUE;
        }
    }
    return FALSE;
}

static void getEditFindText(HWND hEdit)
{
    g_QSearchDlg.szFindTextW[0] = 0;
    if ( g_Plugin.bOldWindows )
        GetWindowTextA( hEdit, (LPSTR) g_QSearchDlg.szFindTextW, MAX_TEXT_SIZE - 1 );
    else
        GetWindowTextW( hEdit, (LPWSTR) g_QSearchDlg.szFindTextW, MAX_TEXT_SIZE - 1 );
}

static void cutEditText(HWND hEdit, BOOL bCutAfterCaret)
{
    DWORD   dwSelPos1 = 0;
    DWORD   dwSelPos2 = 0;
    UINT    len = 0;

    SendMessage( hEdit, EM_GETSEL, (WPARAM) &dwSelPos1, (LPARAM) &dwSelPos2 );

    if ( g_Plugin.bOldWindows )
        len = (UINT) GetWindowTextLengthA(hEdit);
    else
        len = (UINT) GetWindowTextLengthW(hEdit);

    if ( bCutAfterCaret )
    {
        if ( dwSelPos1 < len )
        {
            /*if ( dwSelPos1 < 0 )
                dwSelPos1 = 0;*/
            
            //SendMessage( hEdit, WM_SETREDRAW, FALSE, 0 );
            SendMessage( hEdit, EM_SETSEL, dwSelPos1, -1 );
            //SendMessage( hEdit, WM_SETREDRAW, TRUE, 0 );
            
            if ( g_Plugin.bOldWindows )
            {
                SendMessageA( hEdit, EM_REPLACESEL, TRUE, (LPARAM) "" );
            }
            else
            {
                SendMessageW( hEdit, EM_REPLACESEL, TRUE, (LPARAM) L"" );
            }

            SendMessage( hEdit, EM_SETSEL, dwSelPos1, dwSelPos1 );
        }
    }
    else
    {
        if ( dwSelPos2 > 0 )
        {
            if ( dwSelPos2 > len )
                dwSelPos2 = len;
            
            //SendMessage( hEdit, WM_SETREDRAW, FALSE, 0 );
            SendMessage( hEdit, EM_SETSEL, 0, dwSelPos2 );
            //SendMessage( hEdit, WM_SETREDRAW, TRUE, 0 );
            
            if ( g_Plugin.bOldWindows )
            {
                SendMessageA( hEdit, EM_REPLACESEL, TRUE, (LPARAM) "" );
            }
            else
            {
                SendMessageW( hEdit, EM_REPLACESEL, TRUE, (LPARAM) L"" );
            }

            SendMessage( hEdit, EM_SETSEL, 0, 0 );
        }
    }
}

static BOOL getAkelPadSelectedText(void)
{
    EDITINFO  ei;

    ei.hWndEdit = NULL;
    SendMessage( g_Plugin.hMainWnd, 
      AKD_GETEDITINFO, (WPARAM) NULL, (LPARAM) &ei );
    if ( ei.hWndEdit )
    {
        CHARRANGE cr = { 0, 0 };

        SendMessage( ei.hWndEdit, EM_EXGETSEL, 0, (LPARAM) &cr );
        if ( cr.cpMin != cr.cpMax )
        {
            if ( cr.cpMax >= cr.cpMin + MAX_TEXT_SIZE )
                cr.cpMax = cr.cpMin + MAX_TEXT_SIZE - 1;
            
            if ( g_Plugin.bOldWindows )
            {
                TEXTRANGEA trA;

                trA.chrg.cpMin = cr.cpMin;
                trA.chrg.cpMax = cr.cpMax;
                trA.lpstrText = (LPSTR) g_QSearchDlg.szFindTextW;
                SendMessageA( ei.hWndEdit, EM_GETTEXTRANGE, 0, (LPARAM) &trA );
                if ( g_Options.dwFlags[OPTF_SRCH_USE_SPECIALCHARS] )
                {
                    int  i;
                    int  k;
                    char szTextA[2*MAX_TEXT_SIZE];
                    
                    i = 0;
                    k = 0;
                    while ( trA.lpstrText[i] )
                    {
                        switch ( trA.lpstrText[i] )
                        {
                            case '\t':
                                szTextA[k++] = '\\';
                                szTextA[k++] = 't';
                                break;
                            case '\r':
                                szTextA[k++] = '\\';
                                szTextA[k++] = 'n';
                                break;
                            case '\\':
                            case '*':
                            case '?':
                                szTextA[k++] = '\\';
                            default:
                                szTextA[k++] = trA.lpstrText[i];
                                break;
                        }
                        ++i;
                    }
                    if ( k > i )
                    {
                        if ( k > MAX_TEXT_SIZE - 1 )
                        {
                            k = MAX_TEXT_SIZE - 1;
                        }
                        szTextA[k] = 0;
                        i = 0;
                        while ( (trA.lpstrText[i] = szTextA[i]) != 0 )
                        {
                            ++i;
                        }
                    }
                }
            }
            else
            {
                TEXTRANGEW trW;

                trW.chrg.cpMin = cr.cpMin;
                trW.chrg.cpMax = cr.cpMax;
                trW.lpstrText = (LPWSTR) g_QSearchDlg.szFindTextW;
                SendMessageW( ei.hWndEdit, EM_GETTEXTRANGE, 0, (LPARAM) &trW );
                if ( g_Options.dwFlags[OPTF_SRCH_USE_SPECIALCHARS] )
                {
                    int     i;
                    int     k;
                    wchar_t szTextW[2*MAX_TEXT_SIZE];
                    
                    i = 0;
                    k = 0;
                    while ( trW.lpstrText[i] )
                    {
                        switch ( trW.lpstrText[i] )
                        {
                            case L'\t':
                                szTextW[k++] = L'\\';
                                szTextW[k++] = L't';
                                break;
                            case L'\r':
                                szTextW[k++] = L'\\';
                                szTextW[k++] = L'n';
                                break;
                            case L'\\':
                            case L'*':
                            case L'?':
                                szTextW[k++] = L'\\';
                            default:
                                szTextW[k++] = trW.lpstrText[i];
                                break;
                        }
                        ++i;
                    }
                    if ( k > i )
                    {
                        if ( k > MAX_TEXT_SIZE - 1 )
                        {
                            k = MAX_TEXT_SIZE - 1;
                        }
                        szTextW[k] = 0;
                        i = 0;
                        while ( (trW.lpstrText[i] = szTextW[i]) != 0 )
                        {
                            ++i;
                        }
                    }
                }
            }
            qs_bEditTextChanged = TRUE;
            return TRUE;
        }
    }
    return FALSE;
}

static BOOL EnableDlgItem(HWND hDlg, int itemID, BOOL bEnable)
{
    HWND hDlgItem = GetDlgItem(hDlg, itemID);
    return ( hDlgItem ? EnableWindow(hDlgItem, bEnable) : FALSE );
}

static WNDPROC setWndProc(HWND hWnd, WNDPROC newWndProc)
{
    if ( g_Plugin.bOldWindows )
        return (WNDPROC) SetWindowLongPtrA( hWnd, GWLP_WNDPROC, (LONG_PTR) newWndProc );
    else
        return (WNDPROC) SetWindowLongPtrW( hWnd, GWLP_WNDPROC, (LONG_PTR) newWndProc );
}

static LRESULT callWndProc(WNDPROC prevWndProc, 
  HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    if ( g_Plugin.bOldWindows )
        return CallWindowProcA(prevWndProc, hWnd, uMsg, wParam, lParam);
    else
        return CallWindowProcW(prevWndProc, hWnd, uMsg, wParam, lParam);
}

LRESULT CALLBACK btnWndProc(HWND hBtn,
  UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch ( uMsg )
    {
        case WM_MOUSEMOVE:
        {
            if ( !qs_bBtnTrackMouse )
            {
                TRACKMOUSEEVENT tme;
                DWORD           dwWndStyle;

                tme.cbSize = sizeof(TRACKMOUSEEVENT);
                tme.dwFlags = TME_LEAVE;
                tme.hwndTrack = hBtn;
                tme.dwHoverTime = HOVER_DEFAULT;
                if ( _TrackMouseEvent(&tme) )
                {
                    qs_bBtnTrackMouse = TRUE;
                }
                dwWndStyle = GetWindowLong(hBtn, GWL_STYLE);
                if ( dwWndStyle & BS_FLAT )
                {
                    dwWndStyle -= BS_FLAT;
                    dwWndStyle |= BS_PUSHLIKE;
                    SetWindowLong(hBtn, GWL_STYLE, dwWndStyle);
                    SendMessage( 
                      hBtn, 
                      BM_SETSTYLE, 
                      BS_PUSHLIKE | BS_PUSHBUTTON | BS_CENTER | BS_VCENTER,
                      TRUE );
                }
            }
            break;
        }
        case WM_MOUSELEAVE:
        {
            DWORD dwWndStyle = GetWindowLong(hBtn, GWL_STYLE);
            if ( dwWndStyle & BS_PUSHLIKE )
            {
                dwWndStyle -= BS_PUSHLIKE;
                dwWndStyle |= BS_FLAT;
                SetWindowLong(hBtn, GWL_STYLE, dwWndStyle);
                SendMessage( 
                  hBtn, 
                  BM_SETSTYLE, 
                  BS_FLAT | BS_PUSHBUTTON | BS_CENTER | BS_VCENTER,
                  TRUE );
            }
            qs_bBtnTrackMouse = FALSE;
            break;
        }
        default:
            break;
    }
    
    return callWndProc(prev_btnWndProc, hBtn, uMsg, wParam, lParam);
}

void qsChangeCkeckBoxState(WORD idCheckBox)
{
    const WPARAM wp = (BN_CLICKED << 16) & 0xFFFF0000L;
    WPARAM wState;
    HWND   hCh;
    
    hCh = GetDlgItem(g_QSearchDlg.hDlg, idCheckBox);
    
    if ( idCheckBox == IDC_CH_WHOLEWORD )
    {
        if ( !IsWindowVisible(hCh) )
            return;
    }

    if ( SendMessage(hCh, BM_GETCHECK, 0, 0) == BST_CHECKED )
        wState = BST_UNCHECKED;
    else
        wState = BST_CHECKED;

    SendMessage( hCh, BM_SETCHECK, wState, 0 );
    SendMessage( g_QSearchDlg.hDlg, WM_COMMAND, wp | idCheckBox, (LPARAM) hCh );
}

LRESULT CALLBACK editWndProc(HWND hEdit,
  UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    static DWORD dwSelPos1 = 0;
    static DWORD dwSelPos2 = (DWORD)(-1L);
    static BOOL  bHotKeyPressed = FALSE;
    
    switch ( uMsg )
    {
        case WM_KEYDOWN:
        {
            bHotKeyPressed = FALSE;
            if ( (wParam == VK_RETURN) || (wParam == VK_F3) )
            {
                if ( GetKeyState(VK_QS_PICKUPTEXT) & 0x80 )
                {
                    if ( getAkelPadSelectedText() )
                    {
                        if ( g_Plugin.bOldWindows )
                        {
                            SetWindowTextA( hEdit, (LPCSTR) g_QSearchDlg.szFindTextW );
                        }
                        else
                        {
                            SetWindowTextW( hEdit, (LPCWSTR) g_QSearchDlg.szFindTextW );
                        }
                        SendMessage( g_QSearchDlg.hDlg, QSM_SETNOTFOUND, FALSE, 0 );
                    }
                    SendMessage(hEdit, EM_SETSEL, 0, -1);
                    qs_bEditTextChanged = TRUE;
                    qsearchDoTryHighlightAll(g_QSearchDlg.hDlg);
                }
                else
                {
                    SendMessage( g_QSearchDlg.hDlg, WM_COMMAND, IDOK, 0 );
                }
                return 0;
            }
            else if ( (wParam == VK_DELETE) || (wParam == VK_BACK) )
            {
                qs_bEditTextChanged = TRUE;

                if ( GetKeyState(VK_CONTROL) & 0x80 ) // Ctrl+Del, Ctrl+BS
                {
                    cutEditText( hEdit, (wParam == VK_DELETE) );
                    if ( g_Options.dwFlags[OPTF_SRCH_ONTHEFLY_MODE] )
                    {
                        SendMessage( g_QSearchDlg.hDlg, QSM_FINDFIRST, QSEARCH_NOFINDUP | QSEARCH_NOFINDBEGIN, 0 );
                    }
                    else
                    {
                        SendMessage( g_QSearchDlg.hDlg, QSM_SETNOTFOUND, FALSE, 0 );
                    }
                    return 0;
                }
                else if ( wParam == VK_DELETE ) // Del
                {
                    if ( g_Options.dwFlags[OPTF_SRCH_ONTHEFLY_MODE] )
                    {
                        LRESULT lResult = callWndProc(prev_editWndProc, 
                                            hEdit, uMsg, wParam, lParam);
                        SendMessage( g_QSearchDlg.hDlg, QSM_FINDFIRST, QSEARCH_NOFINDUP | QSEARCH_NOFINDBEGIN, 0 );
                        return lResult;
                    }
                    else
                    {
                        LRESULT lResult = callWndProc(prev_editWndProc, 
                                            hEdit, uMsg, wParam, lParam);
                        SendMessage( g_QSearchDlg.hDlg, QSM_SETNOTFOUND, FALSE, 0 );
                        return lResult;
                    }
                }
                // BS is processed in WM_CHAR
            }
            else if ( wParam == VK_ESCAPE )
            {
                SendMessage( g_QSearchDlg.hDlg, QSM_SHOWHIDE, FALSE, 0 );
                return 0;
            }
            else if ( wParam == VK_TAB )
            {
                return 0;
            }
            else if ( wParam == 0x41 ) // 'A'
            {
                if ( GetKeyState(VK_CONTROL) & 0x80 ) // Ctrl+A
                {
                    SendMessage( hEdit, EM_SETSEL, 0, -1 );
                    return 0;
                }
            }
            else if ( LOBYTE(wParam) == LOBYTE(qs_dwHotKey) )
            {
                bHotKeyPressed = isQSearchHotKeyPressed();
                return 0;
            }
            break;
        }
        case WM_SYSKEYDOWN:
        {
            // it's required when qs_dwHotKey contains Alt
            // (Alt+key do not come as WM_KEYDOWN)
            if ( LOBYTE(wParam) == LOBYTE(qs_dwHotKey) )
            {
                bHotKeyPressed = isQSearchHotKeyPressed();
                return 0;
            }
            if ( (wParam == VK_RETURN) || (wParam == VK_F3) )
            {
                // (if Alt+Enter or Alt+F3 pressed)
                SendMessage(hEdit, WM_KEYDOWN, wParam, lParam);
                return 0;
            }
            /*
            if ( g_Options.dwUseAltHotkeys )
            {
                if ( wParam == g_Options.dwAltMatchCase )
                {
                    qsChangeCkeckBoxState(IDC_CH_MATCHCASE);
                    return 1;
                }
                if ( wParam == g_Options.dwAltHighlightAll )
                {
                    qsChangeCkeckBoxState(IDC_CH_HIGHLIGHTALL);
                    return 1;
                }
                if ( wParam == g_Options.dwAltWholeWord )
                {
                    qsChangeCkeckBoxState(IDC_CH_WHOLEWORD);
                    return 1;
                }
            }
            */
            break;
        }
        case WM_KEYUP:
        {
            if ( bHotKeyPressed )
            {
                bHotKeyPressed = FALSE;
                if ( !qs_bHotKeyPressedOnShow )
                {
                    if ( g_Options.dwFlags[OPTF_HOTKEY_CLOSES_WND] )
                    {
                        SendMessage( g_QSearchDlg.hDlg, QSM_SHOWHIDE, FALSE, 0 );
                    }
                    else
                    {
                        SetFocus( g_Plugin.hMainWnd );
                    }
                    return 0;
                }
            }
            
            qs_bHotKeyPressedOnShow = FALSE;
            
            if ( (wParam == VK_RETURN) || 
                 (wParam == VK_ESCAPE) || 
                 (wParam == VK_TAB) )
            {
                return 0;
            }
            break;
        }
        case WM_SYSKEYUP:
        {
            if ( bHotKeyPressed )
            {
                SendMessage(hEdit, WM_KEYUP, wParam, lParam);
                return 0;
            }
            break;
        }
        case WM_CHAR:
        {
            if ( bHotKeyPressed ||
                 (wParam == VK_RETURN) ||
                 (wParam == VK_ESCAPE) ||
                 (wParam == VK_TAB) ||
                 (wParam == 0x0A) ||      /* 0x0A is Ctrl+Enter. Why? Ask M$ ;) */
                 (wParam == 0x7F) )       /* 0x7F is Ctrl+Backspace. Why? Ask M$ */
            {
                return 0;
            }
            else
            {
                qs_bEditTextChanged = TRUE;
                if ( g_Options.dwFlags[OPTF_SRCH_ONTHEFLY_MODE] )
                {
                    LRESULT lResult = callWndProc(prev_editWndProc, 
                                        hEdit, uMsg, wParam, lParam);
                    SendMessage( g_QSearchDlg.hDlg, QSM_FINDFIRST, QSEARCH_NOFINDUP | QSEARCH_NOFINDBEGIN, 0 );
                    return lResult;
                }
                else
                {
                    LRESULT lResult = callWndProc(prev_editWndProc, 
                                        hEdit, uMsg, wParam, lParam);
                    SendMessage( g_QSearchDlg.hDlg, QSM_SETNOTFOUND, FALSE, 0 );
                    return lResult;
                }
            }
            break;
        }
        case WM_SETFOCUS:
        {
            qs_bEditIsActive = TRUE;
            InvalidateRect(hEdit, NULL, TRUE);
            UpdateWindow(hEdit);
            if ( qs_bEditCanBeNonActive )
            {
                if ( g_Options.dwFlags[OPTF_EDIT_FOCUS_SELECTALL] )
                {
                    SendMessage(hEdit, EM_SETSEL, 0, -1);
                }
                else 
                {
                    if ( !qs_bEditSelJustChanged )
                    {
                        SendMessage(hEdit, EM_SETSEL, dwSelPos1, dwSelPos2);
                    }
                    else
                    {
                        qs_bEditSelJustChanged = FALSE;
                        SendMessage( hEdit, EM_GETSEL, 
                          (WPARAM) &dwSelPos1, (LPARAM) &dwSelPos2 );
                    }
                }
            }
            break;
        }
        case WM_KILLFOCUS:
        {
            if ( qs_bEditCanBeNonActive )
            {
                qs_bEditIsActive = FALSE;
                qs_nEditEOF = 0;
                SendMessage( g_QSearchDlg.hDlg, QSM_SETNOTFOUND, FALSE, 0 );
                /*
                InvalidateRect(hEdit, NULL, TRUE);
                UpdateWindow(hEdit);
                */
                if ( !g_Options.dwFlags[OPTF_EDIT_FOCUS_SELECTALL] )
                {
                    SendMessage( hEdit, EM_GETSEL, 
                      (WPARAM) &dwSelPos1, (LPARAM) &dwSelPos2 );
                }
            }
            break;
        }
        case WM_MOUSEMOVE:
        {
            if ( (!qs_bEditIsActive) && (!qs_bEditTrackMouse) )
            {
                SetFocus(hEdit);
            }
            if ( !qs_bEditTrackMouse )
            {
                TRACKMOUSEEVENT tme;

                tme.cbSize = sizeof(TRACKMOUSEEVENT);
                tme.dwFlags = TME_LEAVE;
                tme.hwndTrack = hEdit;
                tme.dwHoverTime = HOVER_DEFAULT;
                if ( _TrackMouseEvent(&tme) )
                {
                    qs_bEditTrackMouse = TRUE;
                }
            }
            break;
        }
        case WM_MOUSELEAVE:
        {
            qs_bEditTrackMouse = FALSE;
            break;
        }
        case WM_GETDLGCODE:
        {
            if ( !g_Options.dwFlags[OPTF_EDIT_FOCUS_SELECTALL] )
            {
                LRESULT lResult = callWndProc(prev_editWndProc, hEdit, uMsg, wParam, lParam);
                if ( lResult & DLGC_HASSETSEL )
                {
                    lResult -= DLGC_HASSETSEL;
                }
                return lResult;
            }
            break;
        }
        case WM_NOTIFY:
        {
            if ( g_Plugin.bOldWindows )
            {
                if ( ((LPNMHDR) lParam)->code == TTN_GETDISPINFOA )
                {
                    LPNMTTDISPINFOA lpnmdiA = (LPNMTTDISPINFOA) lParam;
                    SendMessage(lpnmdiA->hdr.hwndFrom, TTM_SETMAXTIPWIDTH, 0, 300);
                    lpnmdiA->lpszText = (LPSTR) qsearchGetHintA(IDC_ED_TEXT);
                    return 0;
                }
            }
            else
            {
                if ( ((LPNMHDR) lParam)->code == TTN_GETDISPINFOW )
                {
                    LPNMTTDISPINFOW lpnmdiW = (LPNMTTDISPINFOW) lParam;
                    SendMessage(lpnmdiW->hdr.hwndFrom, TTM_SETMAXTIPWIDTH, 0, 300);
                    lpnmdiW->lpszText = (LPWSTR) qsearchGetHintW(IDC_ED_TEXT);
                    return 0;
                }
            }
            break;
        }
        default:
            break;
    }

    return callWndProc(prev_editWndProc, hEdit, uMsg, wParam, lParam);
}

static BOOL isHighlightMainActive(void)
{
    if ( g_Plugin.bOldWindows )
    {
        PLUGINFUNCTIONA *pfA = (PLUGINFUNCTIONA *) SendMessageA( g_Plugin.hMainWnd,
            AKD_DLLFIND, (WPARAM) cszHighlightMainA, 0 );

        if ( pfA && pfA->bRunning )
        {
            return TRUE;
        }
    }
    else
    {
        PLUGINFUNCTIONW *pfW = (PLUGINFUNCTIONW *) SendMessageW( g_Plugin.hMainWnd,
            AKD_DLLFIND, (WPARAM) cszHighlightMainW, 0 );

        if ( pfW && pfW->bRunning )
        {
            return TRUE;
        }
    }

    return FALSE;
}

static void qsUpdateHighlight(HWND hDlg, HWND hEdit)
{
    wchar_t szFindTextW[MAX_TEXT_SIZE];

    getEditFindText(hEdit);
    
    if ( g_Plugin.bOldWindows )
        lstrcpyA( (LPSTR) szFindTextW, (LPCSTR) g_QSearchDlg.szFindTextW );
    else
        lstrcpyW( szFindTextW, g_QSearchDlg.szFindTextW );

    if ( getAkelPadSelectedText() )
    {
        BOOL    bEqual;
        HWND    hChMatchCase;
        wchar_t szText1[MAX_TEXT_SIZE];
        wchar_t szText2[MAX_TEXT_SIZE];
        const wchar_t* psz1;
        const wchar_t* psz2;

        bEqual = FALSE;
        hChMatchCase = GetDlgItem(hDlg, IDC_CH_MATCHCASE);
        if ( SendMessage(hChMatchCase, BM_GETCHECK, 0, 0) == BST_CHECKED )
        {
            // match case
            psz1 = szFindTextW;
            psz2 = g_QSearchDlg.szFindTextW;
        }
        else
        {
            // case-insensitive
            if ( g_Plugin.bOldWindows )
            {
                lstrcpyA( (LPSTR) szText1, (LPCSTR) szFindTextW );
                CharUpperA( (LPSTR) szText1 );
                lstrcpyA( (LPSTR) szText2, (LPCSTR) g_QSearchDlg.szFindTextW );
                CharUpperA( (LPSTR) szText2 );
            }
            else
            {
                lstrcpyW( szText1, szFindTextW );
                CharUpperW( szText1 );
                lstrcpyW( szText2, g_QSearchDlg.szFindTextW );
                CharUpperW( szText2 );
            }
            psz1 = szText1;
            psz2 = szText2;
        }

        if ( g_Options.dwFlags[OPTF_SRCH_USE_SPECIALCHARS] )
        {
            if ( g_Plugin.bOldWindows )
                bEqual = (match_mask((LPCSTR)psz1, (LPCSTR)psz2, 0) != 0);
            else
                bEqual = (match_maskw(psz1, psz2, 0) != 0);
        }
        else
        {
            if ( g_Plugin.bOldWindows )
                bEqual = (lstrcmpA((LPCSTR)psz1, (LPCSTR)psz2) == 0);
            else
                bEqual = (lstrcmpW(psz1, psz2) == 0);
        }

        if ( g_Plugin.bOldWindows )
            lstrcpyA( (LPSTR) g_QSearchDlg.szFindTextW, (LPCSTR) szFindTextW );
        else
            lstrcpyW( g_QSearchDlg.szFindTextW, szFindTextW );

        qs_bEditTextChanged = FALSE;

        if ( !bEqual )
            qsearchDoSearchText( hEdit, QSEARCH_NEXT );

        qsearchDoTryHighlightAll(hDlg);
    }
    else
    {
        // no text selected - trying to find & highlight
        qsearchDoSearchText( hEdit, QSEARCH_NEXT );
        qsearchDoTryHighlightAll(hDlg);
    }
}

INT_PTR CALLBACK qsearchDlgProc(HWND hDlg,
  UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    static HWND   hEdit = NULL;
    static HWND   hToolTip = NULL;
    static HMENU  hPopupMenu = NULL;
    static HBRUSH hTextNotFoundBrush = NULL;
    static HBRUSH hTextEOFBrush = NULL;
    
    switch ( uMsg )
    {
        case WM_COMMAND:
        {
            unsigned int id = LOWORD(wParam);
            if ( id == IDOK )
            {
                // Originally we get here when Enter is pressed. Why? Ask M$ ;)
                unsigned int uSearch = QSEARCH_NEXT;
                if ( qs_bEditTextChanged && 
                     (qs_bMatchCaseJustChanged ||
                      g_Options.dwFlags[OPTF_SRCH_FROM_BEGINNING] ||
                      !g_Options.dwFlags[OPTF_SRCH_ONTHEFLY_MODE]) )
                {
                    uSearch = QSEARCH_FIRST;
                    qs_bMatchCaseJustChanged = FALSE;
                }
                qs_bEditTextChanged = FALSE;
                getEditFindText( hEdit );
                qsearchDoSearchText( hEdit, uSearch );
                if ( uSearch == QSEARCH_FIRST )
                {
                    qsearchDoTryHighlightAll( hDlg );
                }
                return 1;
            }
            else if ( id == IDCANCEL )
            {
                // We get here when Esc is pressed. Why? Ask M$ ;)
                //qsearchDoQuit( hEdit, hToolTip, hTextNotFoundBrush, hTextEOFBrush );
                qsearchDoShowHide(hDlg, FALSE);
                return 1; 
            }
            else if ( id == IDC_CH_MATCHCASE )
            {
                if ( HIWORD(wParam) == BN_CLICKED )
                {
                    HWND hChHighlightAll;
                    
                    qs_bMatchCaseJustChanged = TRUE;
                    qs_bEditTextChanged = TRUE;
                    qsearchDoSetNotFound(hEdit, FALSE, FALSE);

                    hChHighlightAll = GetDlgItem(hDlg, IDC_CH_HIGHLIGHTALL);
                    if ( SendMessage(hChHighlightAll, BM_GETCHECK, 0, 0) == BST_CHECKED )
                        qsUpdateHighlight(hDlg, hEdit);
                }
            }
            else if ( id == IDC_CH_WHOLEWORD )
            {
                if ( HIWORD(wParam) == BN_CLICKED )
                {
                    HWND hChHighlightAll;

                    qs_bMatchCaseJustChanged = TRUE;
                    qs_bEditTextChanged = TRUE;
                    qsearchDoSetNotFound(hEdit, FALSE, FALSE);

                    hChHighlightAll = GetDlgItem(hDlg, IDC_CH_HIGHLIGHTALL);
                    if ( SendMessage(hChHighlightAll, BM_GETCHECK, 0, 0) == BST_CHECKED )
                    {
                        DWORD dwOnTheFlyMode;

                        getEditFindText(hEdit);
                        dwOnTheFlyMode = g_Options.dwFlags[OPTF_SRCH_ONTHEFLY_MODE];
                        g_Options.dwFlags[OPTF_SRCH_ONTHEFLY_MODE] = 1;
                        qsearchDoSearchText( hEdit, QSEARCH_FIRST );
                        g_Options.dwFlags[OPTF_SRCH_ONTHEFLY_MODE] = dwOnTheFlyMode;
                        qsearchDoTryHighlightAll(hDlg);
                    }
                }
            }
            else if ( id == IDC_CH_HIGHLIGHTALL )
            {
                if ( HIWORD(wParam) == BN_CLICKED )
                {
                    HWND hChHighlightAll;
                    
                    hChHighlightAll = GetDlgItem(hDlg, IDC_CH_HIGHLIGHTALL);
                    if ( SendMessage(hChHighlightAll, BM_GETCHECK, 0, 0) == BST_CHECKED )
                    {
                        g_Options.dwHighlightState |= HLS_IS_CHECKED;
                        qsUpdateHighlight(hDlg, hEdit);
                    }
                    else
                    {
                        if ( (g_Options.dwHighlightState & HLS_IS_CHECKED) == HLS_IS_CHECKED )
                            g_Options.dwHighlightState -= HLS_IS_CHECKED;
                        qsearchDoTryUnhighlightAll();
                    }
                }
            }
            else if ( (id >= IDM_START) && (id < IDM_START + OPTF_COUNT - 1) )
            {
                if ( hPopupMenu )
                {
                    unsigned int state = GetMenuState(hPopupMenu, id, MF_BYCOMMAND);
                    g_Options.dwFlags[id - IDM_START] = 
                        ( (state & MF_CHECKED) == MF_CHECKED ) ? 0 : 1;
                    if ( id == IDM_SRCHONTHEFLYMODE )
                    {
                        /*if ( g_Options.dwFlags[OPTF_SRCH_ONTHEFLY_MODE] )
                        {
                            CheckMenuItem( hPopupMenu, IDM_SRCHUSESPECIALCHARS, 
                              MF_BYCOMMAND | MF_UNCHECKED );
                            g_Options.dwFlags[OPTF_SRCH_USE_SPECIALCHARS] = 0;
                        }*/
                        qsdlgShowHideWholeWordCheckBox(hDlg);
                        qs_nEditEOF = 0;
                        qsearchDoSetNotFound( hEdit, FALSE, FALSE );
                    }
                    else if ( id == IDM_SRCHUSESPECIALCHARS )
                    {
                        /*if ( g_Options.dwFlags[OPTF_SRCH_USE_SPECIALCHARS] )
                        {
                            CheckMenuItem( hPopupMenu, IDM_SRCHONTHEFLYMODE, 
                              MF_BYCOMMAND | MF_UNCHECKED );
                            g_Options.dwFlags[OPTF_SRCH_ONTHEFLY_MODE] = 0;
                        }*/
                        qsdlgShowHideWholeWordCheckBox(hDlg);
                        qs_nEditEOF = 0;
                        qsearchDoSetNotFound( hEdit, FALSE, FALSE );
                    }
                    else if ( id == IDM_SRCHWNDDOCKEDTOP )
                    {
                        if ( g_QSearchDlg.pDockData )
                        {
                            state = DK_SETBOTTOM;
                            if ( g_Options.dwFlags[OPTF_SRCH_WND_DOCKEDTOP] )
                            {
                                state = DK_SETTOP;
                            }
                            SendMessage( g_Plugin.hMainWnd, AKD_DOCK, 
                              state, (LPARAM) g_QSearchDlg.pDockData );
                            PostMessage( g_Plugin.hMainWnd, AKD_RESIZE, 0, 0 );
                        }
                    }
                }
            }
            break;
        }
        case WM_SYSKEYDOWN:
        {
            if ( g_Options.dwUseAltHotkeys )
            {
                if ( wParam == g_Options.dwAltMatchCase )
                {
                    qsChangeCkeckBoxState(IDC_CH_MATCHCASE);
                    return 1;
                }
                if ( wParam == g_Options.dwAltHighlightAll )
                {
                    qsChangeCkeckBoxState(IDC_CH_HIGHLIGHTALL);
                    return 1;
                }
                if ( wParam == g_Options.dwAltWholeWord )
                {
                    qsChangeCkeckBoxState(IDC_CH_WHOLEWORD);
                    return 1;
                }
            }
            break;
        }
        case WM_SETFOCUS:
        {
            EnableWindow( 
              GetDlgItem(hDlg, IDC_CH_HIGHLIGHTALL), 
              g_bHighlightPlugin
            );
            
            if ( !hEdit )
            {
                hEdit = GetDlgItem(hDlg, IDC_ED_TEXT);
            }
            qs_bEditIsActive = TRUE;
            SetFocus(hEdit);
            break;
        }
        case WM_CTLCOLOREDIT:
        {
            if ( hEdit == (HWND) lParam )
            {
                int nBkColor = COLOR_BTNFACE; // normal inactive
                if ( qs_bEditIsActive )
                {  
                    if ( !qs_bEditNotFound )
                    {
                        if ( (!qs_bEditIsEOF) || (g_Options.colorEOF == RGB(0xFF,0xFF,0xFF)) )
                            nBkColor = COLOR_WINDOW; // normal active
                        else
                            nBkColor = 0x7FFE; // EOF active
                    }
                    else
                        nBkColor = 0x7FFF; // not found active
                }
                SetTextColor( (HDC) wParam, GetSysColor(COLOR_WINDOWTEXT) );
                SetBkMode( (HDC) wParam, TRANSPARENT );
                switch ( nBkColor )
                {
                    case 0x7FFF:
                        SetBkColor( (HDC) wParam, g_Options.colorNotFound );
                        return (LRESULT) hTextNotFoundBrush;

                    case 0x7FFE:
                        SetBkColor( (HDC) wParam, g_Options.colorEOF );
                        return (LRESULT) hTextEOFBrush;

                    default:
                        SetBkColor( (HDC) wParam, GetSysColor(nBkColor) );
                        return (LRESULT) GetSysColorBrush(nBkColor);
                }
            }
            break;
        }
        case WM_SHOWWINDOW:
        {
            if ( wParam )
            {
                HWND hChHighlightAll;
                
                g_bHighlightPlugin = isHighlightMainActive();
                
                hChHighlightAll = GetDlgItem(hDlg, IDC_CH_HIGHLIGHTALL);
                EnableWindow( hChHighlightAll, g_bHighlightPlugin );
                SendMessage( hChHighlightAll, BM_SETCHECK, 
                  (((g_Options.dwHighlightState & HLS_IS_CHECKED) == HLS_IS_CHECKED) ? 
                       BST_CHECKED : BST_UNCHECKED), 0 );
                
                qs_dwHotKey = getQSearchHotKey();
                qs_bHotKeyPressedOnShow = isQSearchHotKeyPressed();

                SetFocus(hEdit);
            }
            break;
        }
        case WM_SIZE:
        {
            if ( wParam != SIZE_MINIMIZED )
            {
                if ( g_QSearchDlg.pDockData )
                {
                    g_QSearchDlg.pDockData->rcDragDrop.left = 0;
                    g_QSearchDlg.pDockData->rcDragDrop.top = 0;
                    g_QSearchDlg.pDockData->rcDragDrop.right = LOWORD(lParam);
                    g_QSearchDlg.pDockData->rcDragDrop.bottom = HIWORD(lParam);
                }
                return 1;
            }
            break;
        }
        case WM_CONTEXTMENU:
        {
            int   i;
            POINT pt;

            pt.x = LOWORD(lParam);
            pt.y = HIWORD(lParam);

            for ( i = 0; i < OPTF_COUNT - 1; i++ )
            {
                CheckMenuItem( hPopupMenu, IDM_START + i, 
                  MF_BYCOMMAND | (g_Options.dwFlags[i] ? MF_CHECKED : MF_UNCHECKED) );
            }
        
            TrackPopupMenuEx(hPopupMenu, 0, pt.x, pt.y, hDlg, NULL);
            
            break;
        }
        case QSM_FINDNEXT:
        {
            getEditFindText( hEdit );
            if ( qs_bMatchCaseJustChanged )
            {
                qs_bMatchCaseJustChanged = FALSE;
                qsearchDoSearchText( hEdit, QSEARCH_FIRST );
                qsearchDoTryHighlightAll( hDlg );
            }
            else
            {
                qsearchDoSearchText( hEdit, QSEARCH_NEXT );
            }
            return 1;
        }
        case QSM_FINDFIRST:
        {
            DWORD dwSearch = QSEARCH_FIRST;

            if ( wParam & QSEARCH_NOFINDUP )
                dwSearch |= QSEARCH_NOFINDUP;

            if ( wParam & QSEARCH_NOFINDBEGIN )
                dwSearch |= QSEARCH_NOFINDBEGIN;
            
            getEditFindText( hEdit );
            qsearchDoSearchText( hEdit, dwSearch );
            qsearchDoTryHighlightAll( hDlg );
            return 1;
        }
        case QSM_SELFIND:
        {
            qsearchDoSelFind( hEdit, (BOOL) wParam );
            return 1;
        }
        case QSM_GETHWNDEDIT:
        {
            if ( lParam )
                *((HWND *)lParam) = hEdit;
            return 1;
        }
        case QSM_SHOWHIDE:
        {
            qsearchDoShowHide( hDlg, (BOOL) wParam );
            return 1;
        }
        case QSM_SETNOTFOUND:
        {
            qsearchDoSetNotFound( hEdit, (BOOL) wParam, FALSE );
            return 1;
        }
        case QSM_QUIT:
        {
            qsearchDoQuit( hEdit, hToolTip, hTextNotFoundBrush, hTextEOFBrush );
            return 1;
        }
        case WM_INITDIALOG:
        {
            HWND hDlgItm = GetDlgItem(hDlg, IDC_CH_MATCHCASE);
            if ( hDlgItm )
            {
                SendMessage( hDlgItm, BM_SETCHECK, 
                  g_QSearchDlg.bMatchCase ? BST_CHECKED : BST_UNCHECKED, 0 );
            }
            g_QSearchDlg.hDlg = hDlg;
        
            // restoring initial values of static vars
            qs_bEditIsActive = TRUE;
            qs_bEditCanBeNonActive = TRUE;
            qs_bEditSelJustChanged = FALSE;
            qs_bEditNotFound = FALSE;
            qs_bEditIsEOF = FALSE;
            qs_nEditEOF = 0;
            qs_bEditTextChanged = TRUE;
            qs_bEditTrackMouse = FALSE;
            qs_bBtnTrackMouse = FALSE;
            qs_bHotKeyPressedOnShow = FALSE;
            qs_dwHotKey = 0;
        
            hTextNotFoundBrush = CreateSolidBrush(g_Options.colorNotFound);
            if ( g_Options.colorEOF != RGB(0xFF,0xFF,0xFF) )
            {
                hTextEOFBrush = CreateSolidBrush(g_Options.colorEOF);
            }
            hEdit = GetDlgItem(hDlg, IDC_ED_TEXT);
            SendMessage(hEdit, EM_SETLIMITTEXT, MAX_TEXT_SIZE - 1, 0);
            prev_editWndProc = setWndProc(hEdit, editWndProc);
            hDlgItm = GetDlgItem(hDlg, IDCANCEL);
            if ( hDlgItm )
            {
                prev_btnWndProc = setWndProc(hDlgItm, btnWndProc);
            }
            hToolTip = qsearchDoInitToolTip(hDlg, hEdit);
            if ( g_Plugin.bOldWindows )
            {
                hPopupMenu = LoadMenuA( g_Plugin.hInstanceDLL, 
                  MAKEINTRESOURCEA(IDR_MENU_OPTIONS) );
            }
            else
            {
                hPopupMenu = LoadMenuW( g_Plugin.hInstanceDLL, 
                  MAKEINTRESOURCEW(IDR_MENU_OPTIONS) );
            }
            hPopupMenu = GetSubMenu(hPopupMenu, 0);
            qsearchSetPopupMenuLang(hPopupMenu);
            qsearchSetDialogLang(hDlg);

            qsdlgShowHideWholeWordCheckBox(hDlg);
            
            if ( g_Options.dwFlags[OPTF_SRCH_PICKUP_SELECTION] )
            {
                getAkelPadSelectedText();
            }
            if ( g_Plugin.bOldWindows )
            {
                SetWindowTextA( hEdit, (LPCSTR) g_QSearchDlg.szFindTextW );
            }
            else
            {
                SetWindowTextW( hEdit, (LPCWSTR) g_QSearchDlg.szFindTextW );
            }
            //SetFocus(hEdit);
            if ( !g_Options.dwFlags[OPTF_EDIT_FOCUS_SELECTALL] )
                SendMessage(hEdit, EM_SETSEL, 0, -1);

            break;
        }
        default:
            break;
    }
    return 0;
}

static void fillToolInfoA(
  TOOLINFOA* lptiA,
  LPSTR      lpToolTipText,
  HWND       hWnd)
{
    static unsigned int uId = 0;
    RECT rect;

    GetClientRect(hWnd, &rect);
    
    lptiA->cbSize = sizeof(TOOLINFOA);
    lptiA->uFlags = TTF_SUBCLASS;
    lptiA->hwnd = hWnd;
    lptiA->hinst = g_Plugin.hInstanceDLL;
    lptiA->uId = uId;
    lptiA->lpszText = lpToolTipText;
    // ToolTip control will cover the whole window
    lptiA->rect.left = rect.left;    
    lptiA->rect.top = rect.top;
    lptiA->rect.right = rect.right;
    lptiA->rect.bottom = rect.bottom;
    lptiA->lParam = 0;

    ++uId;
}

static void fillToolInfoW(
  TOOLINFOW* lptiW,
  LPWSTR     lpToolTipText,
  HWND       hWnd)
{
    static unsigned int uId = 0;
    RECT rect;

    GetClientRect(hWnd, &rect);
    
    lptiW->cbSize = sizeof(TOOLINFOW);
    lptiW->uFlags = TTF_SUBCLASS;
    lptiW->hwnd = hWnd;
    lptiW->hinst = g_Plugin.hInstanceDLL;
    lptiW->uId = uId;
    lptiW->lpszText = lpToolTipText;
    // ToolTip control will cover the whole window
    lptiW->rect.left = rect.left;    
    lptiW->rect.top = rect.top;
    lptiW->rect.right = rect.right;
    lptiW->rect.bottom = rect.bottom;
    lptiW->lParam = 0;

    ++uId;
}

HWND qsearchDoInitToolTip(HWND hDlg, HWND hEdit)
{
    INITCOMMONCONTROLSEX iccex;
    HWND                 hToolTip = NULL;

    iccex.dwICC = ICC_WIN95_CLASSES;
    iccex.dwSize = sizeof(INITCOMMONCONTROLSEX);
    InitCommonControlsEx( &iccex );

    if ( g_Plugin.bOldWindows )
    {
        hToolTip = CreateWindowExA( WS_EX_TOPMOST, TOOLTIPS_CLASSA, 
          0, WS_POPUP | TTS_NOPREFIX | TTS_ALWAYSTIP, 
          CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, 
          hDlg, 0, g_Plugin.hInstanceDLL, 0 );
    }
    else
    {
        hToolTip = CreateWindowExW( WS_EX_TOPMOST, TOOLTIPS_CLASSW, 
          0, WS_POPUP | TTS_NOPREFIX | TTS_ALWAYSTIP, 
          CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, 
          hDlg, 0, g_Plugin.hInstanceDLL, 0 );

    }
    
    if ( hToolTip )
    {
        SetWindowPos( hToolTip, HWND_TOPMOST,
          0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);

        if ( g_Plugin.bOldWindows )
        {
            TOOLINFOA tiA;

            fillToolInfoA( &tiA, LPSTR_TEXTCALLBACKA, hEdit );
            // LPSTR_TEXTCALLBACKA means "send TTN_GETDISPINFOA to hEdit"
            SendMessage( hToolTip, TTM_ADDTOOLA, 0, (LPARAM) &tiA );

            fillToolInfoA( &tiA, (LPSTR) qsearchGetHintA(IDCANCEL), 
              GetDlgItem(hDlg, IDCANCEL) );
            SendMessage( hToolTip, TTM_ADDTOOLA, 0, (LPARAM) &tiA );

            fillToolInfoA( &tiA, (LPSTR) qsearchGetHintA(IDC_CH_MATCHCASE), 
              GetDlgItem(hDlg, IDC_CH_MATCHCASE) );
            SendMessage( hToolTip, TTM_ADDTOOLA, 0, (LPARAM) &tiA );

            fillToolInfoA( &tiA, (LPSTR) qsearchGetHintA(IDC_CH_WHOLEWORD), 
              GetDlgItem(hDlg, IDC_CH_WHOLEWORD) );
            SendMessage( hToolTip, TTM_ADDTOOLA, 0, (LPARAM) &tiA );

            fillToolInfoA( &tiA, (LPSTR) qsearchGetHintA(IDC_CH_HIGHLIGHTALL), 
              GetDlgItem(hDlg, IDC_CH_HIGHLIGHTALL) );
            SendMessage( hToolTip, TTM_ADDTOOLA, 0, (LPARAM) &tiA );
        }
        else
        {
            TOOLINFOW tiW;

            fillToolInfoW( &tiW, LPSTR_TEXTCALLBACKW, hEdit );
            // LPSTR_TEXTCALLBACKW means "send TTN_GETDISPINFOW to hEdit"
            SendMessage( hToolTip, TTM_ADDTOOLW, 0, (LPARAM) &tiW );

            fillToolInfoW( &tiW, (LPWSTR) qsearchGetHintW(IDCANCEL), 
              GetDlgItem(hDlg, IDCANCEL) );
            SendMessage( hToolTip, TTM_ADDTOOLW, 0, (LPARAM) &tiW );

            fillToolInfoW( &tiW, (LPWSTR) qsearchGetHintW(IDC_CH_MATCHCASE), 
              GetDlgItem(hDlg, IDC_CH_MATCHCASE) );
            SendMessage( hToolTip, TTM_ADDTOOLW, 0, (LPARAM) &tiW );

            fillToolInfoW( &tiW, (LPWSTR) qsearchGetHintW(IDC_CH_WHOLEWORD), 
              GetDlgItem(hDlg, IDC_CH_WHOLEWORD) );
            SendMessage( hToolTip, TTM_ADDTOOLW, 0, (LPARAM) &tiW );

            fillToolInfoW( &tiW, (LPWSTR) qsearchGetHintW(IDC_CH_HIGHLIGHTALL), 
              GetDlgItem(hDlg, IDC_CH_HIGHLIGHTALL) );
            SendMessage( hToolTip, TTM_ADDTOOLW, 0, (LPARAM) &tiW );
        }
    }
    
    return hToolTip;
}

void qsearchDoQuit(HWND hEdit, HWND hToolTip, HBRUSH hBrush1, HBRUSH hBrush2)
{
    HWND hDlgItm = GetDlgItem(g_QSearchDlg.hDlg, IDC_CH_MATCHCASE);
    if ( hDlgItm )
    {
        if ( SendMessage(hDlgItm, BM_GETCHECK, 0, 0) == BST_CHECKED )
            g_QSearchDlg.bMatchCase = TRUE;
        else
            g_QSearchDlg.bMatchCase = FALSE;
    }
        
    getEditFindText(hEdit);

    if ( prev_editWndProc )
    {
        setWndProc(hEdit, prev_editWndProc);
        prev_editWndProc = NULL;
    }
    if ( prev_btnWndProc )
    {
        hDlgItm = GetDlgItem(g_QSearchDlg.hDlg, IDCANCEL);
        setWndProc(hDlgItm, prev_btnWndProc);
        prev_btnWndProc = NULL;
    }
    if ( hToolTip )
    {
        DestroyWindow( hToolTip );
    }
    if ( hBrush1 )
    {
        DeleteObject(hBrush1);
    }
    if ( hBrush2 )
    {
        DeleteObject(hBrush2);
    }
    DestroyWindow( g_QSearchDlg.hDlg );
    g_QSearchDlg.hDlg = NULL;
    
    if ( !g_Plugin.bAkelPadOnFinish )
    {
        Uninitialize();
        SendMessage(g_Plugin.hMainWnd, AKD_RESIZE, 0, 0);
        SetFocus(g_Plugin.hMainWnd);
    }
}

void qsearchDoSetNotFound(HWND hEdit, BOOL bNotFound, BOOL bEOF)
{
    if ( bNotFound )
        qsearchDoTryUnhighlightAll();
    
    qs_bEditNotFound = bNotFound;
    qs_bEditIsEOF = bEOF;
    InvalidateRect(hEdit, NULL, TRUE);
    UpdateWindow(hEdit);
}

void qsearchDoShowHide(HWND hDlg, BOOL bShow)
{
    BOOL bChangeSelection = !IsWindowVisible(hDlg);
    
    qsearchDoSetNotFound( GetDlgItem(hDlg, IDC_ED_TEXT), FALSE, FALSE );
    
    if ( bShow )
    {
        HWND hCloseBtn = GetDlgItem(hDlg, IDCANCEL);

        SendMessage(hCloseBtn, BM_SETSTATE, FALSE, 0);
        SendMessage( 
            hCloseBtn, 
            BM_SETSTYLE, 
            BS_FLAT | BS_PUSHBUTTON | BS_CENTER | BS_VCENTER,
            TRUE );

        /*
        // The following line is commented because 
        // the hot key is re-read on WM_SHOWWINDOW.
        qs_dwHotKey = getQSearchHotKey();
        */
        qs_bHotKeyPressedOnShow = isQSearchHotKeyPressed();
    }
    else
    {
        qs_nEditEOF = 0;
    }

    SendMessage( 
        g_Plugin.hMainWnd, 
        AKD_DOCK,
        (bShow ? DK_SHOW : DK_HIDE),
        (LPARAM) g_QSearchDlg.pDockData );

    // Change AkelPad's plugin status (running/not running)
    if ( g_szFunctionQSearchW[0] )
    {
        if ( g_Plugin.bOldWindows )
        {
            PLUGINFUNCTIONA* pfA;
            
            pfA = (PLUGINFUNCTIONA *) SendMessage( g_Plugin.hMainWnd, 
              AKD_DLLFIND, (WPARAM) g_szFunctionQSearchW, 0 );
            if ( pfA )
            {
                pfA->bRunning = bShow;
            }
        }
        else
        {
            PLUGINFUNCTIONW* pfW;

            pfW = (PLUGINFUNCTIONW *) SendMessage( g_Plugin.hMainWnd, 
              AKD_DLLFIND, (WPARAM) g_szFunctionQSearchW, 0 );
            if ( pfW )
            {
                pfW->bRunning = bShow;
            }
        }

        // Send "empty" command to AkelPad to update the Toolbar plugin state
        SendMessage(g_Plugin.hMainWnd, WM_COMMAND, 0, 0);
    }
    
    if ( bShow )
    {
        HWND hEdit = GetDlgItem(hDlg, IDC_ED_TEXT);

        if ( g_Options.dwFlags[OPTF_SRCH_PICKUP_SELECTION] )
        {
            if ( getAkelPadSelectedText() )
            {
                bChangeSelection = TRUE;
            }
            if ( g_Plugin.bOldWindows )
            {
                SetWindowTextA( hEdit, (LPCSTR) g_QSearchDlg.szFindTextW );
            }
            else
            {
                SetWindowTextW( hEdit, (LPCWSTR) g_QSearchDlg.szFindTextW );
            }
        }
        if ( bChangeSelection && !g_Options.dwFlags[OPTF_EDIT_FOCUS_SELECTALL] )
        {
            SendMessage(hEdit, EM_SETSEL, 0, -1);
            qs_bEditSelJustChanged = TRUE;
        }
        SetFocus(hEdit);
    }
    else
    {
        //SendMessage(g_Plugin.hMainWnd, AKD_RESIZE, 0, 0);
        SetFocus(g_Plugin.hMainWnd);
    }
}

void qsearchDoSelFind(HWND hEdit, BOOL bFindPrev)
{
    wchar_t prevFindTextW[MAX_TEXT_SIZE];
    DWORD   prevSrchFromBeginning;
    DWORD   prevSrchUseSpecialChars;

    if ( g_Plugin.bOldWindows )
    {
        lstrcpyA( (LPSTR) prevFindTextW, (LPCSTR) g_QSearchDlg.szFindTextW );
    }
    else
    {
        lstrcpyW( (LPWSTR) prevFindTextW, (LPCWSTR) g_QSearchDlg.szFindTextW );
    }

    if ( g_Options.dwFlags[OPTF_SRCH_SELFIND_PICKUP] )
    {    
        // getting selected text with current search flags
        if ( getAkelPadSelectedText() )
        {
            if ( g_Plugin.bOldWindows )
            {
                SetWindowTextA( hEdit, (LPCSTR) g_QSearchDlg.szFindTextW );
            }
            else
            {
                SetWindowTextW( hEdit, (LPCWSTR) g_QSearchDlg.szFindTextW );
            }
        }
    }
    
    // saving search flags
    prevSrchFromBeginning = g_Options.dwFlags[OPTF_SRCH_FROM_BEGINNING];
    prevSrchUseSpecialChars = g_Options.dwFlags[OPTF_SRCH_USE_SPECIALCHARS];

    // these search flags must be disabled here
    g_Options.dwFlags[OPTF_SRCH_FROM_BEGINNING] = 0;
    g_Options.dwFlags[OPTF_SRCH_USE_SPECIALCHARS] = 0;
    
    // getting selected text with modified search flags
    if ( getAkelPadSelectedText() )
    {
        DWORD   dwSearchParams;
        BOOL    bMatchCase;
        HWND    hDlgItm;

        bMatchCase = FALSE;
        if ( hDlgItm = GetDlgItem(g_QSearchDlg.hDlg, IDC_CH_MATCHCASE) )
        {
            if ( SendMessage(hDlgItm, BM_GETCHECK, 0, 0) == BST_CHECKED )
                bMatchCase = TRUE;
        }

        // clear the "not found" flag
        qs_bEditNotFound = FALSE;
        qs_bEditIsEOF = FALSE;
        
        if ( g_Plugin.bOldWindows )
        {
            int (WINAPI *cmpfuncA)(LPCSTR, LPCSTR) = bMatchCase ? lstrcmpA : lstrcmpiA;
            if ( cmpfuncA((LPCSTR) prevFindTextW, (LPCSTR) g_QSearchDlg.szFindTextW) != 0 )
            {
                qs_bEditTextChanged = TRUE;
                qs_nEditEOF = 0;
            }
        }
        else
        {
            int (WINAPI *cmpfuncW)(LPCWSTR, LPCWSTR) = bMatchCase ? lstrcmpW : lstrcmpiW;
            if ( cmpfuncW((LPCWSTR) prevFindTextW, (LPCWSTR) g_QSearchDlg.szFindTextW) != 0 )
            {
                qs_bEditTextChanged = TRUE;
                qs_nEditEOF = 0;
            }
        }

        dwSearchParams = QSEARCH_NEXT | QSEARCH_SEL;
        if ( bFindPrev )  dwSearchParams |= QSEARCH_SEL_FINDUP;
        qsearchDoSearchText(hEdit, dwSearchParams);
    }

    // restoring search flags
    g_Options.dwFlags[OPTF_SRCH_FROM_BEGINNING] = prevSrchFromBeginning;
    g_Options.dwFlags[OPTF_SRCH_USE_SPECIALCHARS] = prevSrchUseSpecialChars;
}

// searches for g_QSearchDlg.szFindTextW
void qsearchDoSearchText(HWND hEdit, DWORD dwParams)
{
    EDITINFO ei;
    HWND     hDlgItm;
    BOOL     bNotFound = FALSE;
    BOOL     bEOF = FALSE;
    DWORD    dwSearchFlags = FR_DOWN;
    
    if ( (dwParams & QSEARCH_FIRST) == QSEARCH_FIRST )
    {
        // searching for the first time, clear the EOF flag
        qs_nEditEOF = 0;
    }
    else
    {
        // searching for the same text again
        if ( qs_bEditNotFound )
            return;
    }

    if ( g_Plugin.bOldWindows )
    {
        if ( 0 == ((LPCSTR) g_QSearchDlg.szFindTextW)[0] )
        {
            qsearchDoSetNotFound(hEdit, FALSE, FALSE);
            return;
        }
    }
    else
    {
        if ( 0 == ((LPCWSTR) g_QSearchDlg.szFindTextW)[0] )
        {
            qsearchDoSetNotFound(hEdit, FALSE, FALSE);
            return;
        }
    }

    ei.hWndEdit = 0;
    SendMessage( g_Plugin.hMainWnd, AKD_GETEDITINFO, 0, (LPARAM) &ei );
    if ( !ei.hWndEdit )
    {
        return;
    }

    if ( ((dwParams & QSEARCH_SEL) != QSEARCH_SEL) ||
         g_Options.dwFlags[OPTF_SRCH_SELFIND_PICKUP] )  
    {
        // not SelFindNext or SelFindPrev
        // or srch_selfind_pickup = 1
        g_QSearchDlg.bQSearching = TRUE;
    }

    if ( ((dwParams & QSEARCH_SEL_FINDUP) == QSEARCH_SEL_FINDUP) ||
         (((dwParams & QSEARCH_SEL) != QSEARCH_SEL) &&
          ((dwParams & QSEARCH_NOFINDUP) != QSEARCH_NOFINDUP) &&
          ((GetKeyState(VK_QS_FINDUP) & 0x80) == 0x80)) )
    {
        dwSearchFlags = FR_UP;
    }
    
    if ( ((dwParams & QSEARCH_SEL) != QSEARCH_SEL) &&
         ((dwParams & QSEARCH_NOFINDBEGIN) != QSEARCH_NOFINDBEGIN) &&
         ((GetKeyState(VK_QS_FINDBEGIN) & 0x80) == 0x80) )
    {
        dwSearchFlags |= FR_BEGINNING;
    }

    if ( hDlgItm = GetDlgItem(g_QSearchDlg.hDlg, IDC_CH_MATCHCASE) )
    {
        if ( SendMessage(hDlgItm, BM_GETCHECK, 0, 0) == BST_CHECKED )
            dwSearchFlags |= FR_MATCHCASE;
    }

    if ( !g_Options.dwFlags[OPTF_SRCH_USE_SPECIALCHARS] )
    {
        if ( hDlgItm = GetDlgItem(g_QSearchDlg.hDlg, IDC_CH_WHOLEWORD) )
        {
            if ( SendMessage(hDlgItm, BM_GETCHECK, 0, 0) == BST_CHECKED )
                dwSearchFlags |= FR_WHOLEWORD;
        }
    }

    // this forbids deactivating of the Edit control
    // when the text is typed and searched
    qs_bEditCanBeNonActive = FALSE;
        
    if ( (dwParams & QSEARCH_FIRST) == QSEARCH_FIRST )
    {
        if ( g_Options.dwFlags[OPTF_SRCH_ONTHEFLY_MODE] )
        {
            CHARRANGE cr = {0, 0};
            SendMessage( ei.hWndEdit, EM_EXGETSEL, 0, (LPARAM) &cr );
            cr.cpMax = cr.cpMin;
            SendMessage( ei.hWndEdit, EM_EXSETSEL, 0, (LPARAM) &cr );
        }
        if ( g_Options.dwFlags[OPTF_SRCH_FROM_BEGINNING] )
            dwSearchFlags |= FR_BEGINNING;
    }
    
    if ( g_Plugin.bOldWindows )
    {
        TEXTFINDA tfA;
        int       srchEOF;
        BOOL      bSearchEx = FALSE;
        int       iFindResult = -1;
        
        //g_QSearchDlg.szFindTextW[0] = 0;
        //GetWindowTextA( hEdit, (LPSTR) g_QSearchDlg.szFindTextW, MAX_TEXT_SIZE - 1 );
        tfA.dwFlags = dwSearchFlags;
        tfA.pFindIt = (LPSTR) getTextToSearch(g_QSearchDlg.szFindTextW, &bSearchEx);
        srchEOF = (dwSearchFlags & FR_UP) ? QSEARCH_EOF_UP : QSEARCH_EOF_DOWN;
        if ( ((qs_nEditEOF & srchEOF) == 0) || !IsWindowVisible(g_QSearchDlg.hDlg) )
        {
            if ( !bSearchEx )
            {
                iFindResult = (int) SendMessage( g_Plugin.hMainWnd, 
                  AKD_TEXTFIND, (WPARAM) ei.hWndEdit, (LPARAM) &tfA );
            }
            else
            {
                iFindResult = doFindTextExA( ei.hWndEdit, &tfA ) ;
            }
        }
        if ( iFindResult == -1 )
        {
            if ( g_Options.dwFlags[OPTF_SRCH_STOP_EOF] && 
                 ((qs_nEditEOF & srchEOF) == 0) && 
                 ((dwParams & QSEARCH_NEXT) == QSEARCH_NEXT) )
            {
                HWND hWndFocused = GetFocus();
                
                qs_nEditEOF = srchEOF;
                MessageBoxA( 
                  g_Plugin.hMainWnd, 
                  qsearchGetHintA(IDS_EOFREACHED),
                  "AkelPad (QSearch)",
                  MB_OK | MB_ICONASTERISK
                );
                
                if ( IsWindowVisible(g_QSearchDlg.hDlg) && (hWndFocused != ei.hWndEdit) )
                    SetFocus(hEdit);
                else
                    SetFocus(ei.hWndEdit);
            }
            else
            {
                bEOF = TRUE;
                
                if ( (dwSearchFlags & FR_UP) == FR_UP )
                {
                    LONG      pos = 0;
                    CHARRANGE cr = {0, 0};
                    SendMessage( ei.hWndEdit, EM_EXGETSEL, 0, (LPARAM) &cr );
                    pos = cr.cpMin;
                    cr.cpMin = -1;
                    cr.cpMax = -1;
                    SendMessage( ei.hWndEdit, EM_EXSETSEL, 0, (LPARAM) &cr );
                    if ( tfA.dwFlags & FR_BEGINNING )
                    {
                        tfA.dwFlags -= FR_BEGINNING;
                    }
                    if ( !bSearchEx )
                    {
                        iFindResult = (int) SendMessage( g_Plugin.hMainWnd, 
                          AKD_TEXTFIND, (WPARAM) ei.hWndEdit, (LPARAM) &tfA );
                    }
                    else
                    {
                        iFindResult = doFindTextExA( ei.hWndEdit, &tfA ) ;
                    }
                    if ( iFindResult == -1 )
                    {
                        cr.cpMin = pos;
                        cr.cpMax = pos;
                        SendMessage( ei.hWndEdit, EM_EXSETSEL, 0, (LPARAM) &cr );
                        bNotFound = TRUE;
                    }
                }
                else
                {
                    if ( ((dwSearchFlags & FR_BEGINNING) != FR_BEGINNING) ||
                         (qs_nEditEOF & srchEOF) )
                    {
                        tfA.dwFlags = dwSearchFlags | FR_BEGINNING;
                        if ( !bSearchEx )
                        {
                            iFindResult = (int) SendMessage( g_Plugin.hMainWnd, 
                              AKD_TEXTFIND, (WPARAM) ei.hWndEdit, (LPARAM) &tfA );
                        }
                        else
                        {
                            iFindResult = doFindTextExA( ei.hWndEdit, &tfA ) ;
                        }
                        if ( iFindResult == -1 )
                        {
                            bNotFound = TRUE;
                        }
                    }
                    else
                    {
                        bNotFound = TRUE;
                    }
                }
                qs_nEditEOF = 0;
            }
        }
        else
        {
            qs_nEditEOF = 0;
        }
    }
    else
    {
        TEXTFINDW tfW;
        int       srchEOF;
        BOOL      bSearchEx = FALSE;
        int       iFindResult = -1;
        
        //g_QSearchDlg.szFindTextW[0] = 0;
        //GetWindowTextW( hEdit, (LPWSTR) g_QSearchDlg.szFindTextW, MAX_TEXT_SIZE - 1 );
        tfW.dwFlags = dwSearchFlags;
        tfW.pFindIt = (LPWSTR) getTextToSearch(g_QSearchDlg.szFindTextW, &bSearchEx);
        srchEOF = (dwSearchFlags & FR_UP) ? QSEARCH_EOF_UP : QSEARCH_EOF_DOWN;
        if ( ((qs_nEditEOF & srchEOF) == 0) || !IsWindowVisible(g_QSearchDlg.hDlg) )
        {
            if ( !bSearchEx )
            {
                iFindResult = (int) SendMessage( g_Plugin.hMainWnd, 
                  AKD_TEXTFIND, (WPARAM) ei.hWndEdit, (LPARAM) &tfW );
            }
            else       
            {
                iFindResult = doFindTextExW( ei.hWndEdit, &tfW ) ;
            }
        }
        if ( iFindResult == -1 )
        {
            if ( g_Options.dwFlags[OPTF_SRCH_STOP_EOF] && 
                 ((qs_nEditEOF & srchEOF) == 0) &&
                 ((dwParams & QSEARCH_NEXT) == QSEARCH_NEXT) )
            {
                HWND hWndFocused = GetFocus();
                
                qs_nEditEOF = srchEOF;
                MessageBoxW( 
                  g_Plugin.hMainWnd, 
                  qsearchGetHintW(IDS_EOFREACHED),
                  L"AkelPad (QSearch)",
                  MB_OK | MB_ICONASTERISK
                );
                
                if ( IsWindowVisible(g_QSearchDlg.hDlg) && (hWndFocused != ei.hWndEdit) )
                    SetFocus(hEdit);
                else
                    SetFocus(ei.hWndEdit);
            }
            else
            {
                bEOF = TRUE;
                
                if ( (dwSearchFlags & FR_UP) == FR_UP )
                {
                    LONG      pos = 0;
                    CHARRANGE cr = {0, 0};
                    SendMessage( ei.hWndEdit, EM_EXGETSEL, 0, (LPARAM) &cr );
                    pos = cr.cpMin;
                    cr.cpMin = -1;
                    cr.cpMax = -1;
                    SendMessage( ei.hWndEdit, EM_EXSETSEL, 0, (LPARAM) &cr );
                    if ( tfW.dwFlags & FR_BEGINNING )
                    {
                        tfW.dwFlags -= FR_BEGINNING;
                    }
                    if ( !bSearchEx )
                    {
                        iFindResult = (int) SendMessage( g_Plugin.hMainWnd, 
                          AKD_TEXTFIND, (WPARAM) ei.hWndEdit, (LPARAM) &tfW );
                    }
                    else       
                    {
                        iFindResult = doFindTextExW( ei.hWndEdit, &tfW );
                    }
                    if ( iFindResult == -1 )
                    {
                        cr.cpMin = pos;
                        cr.cpMax = pos;
                        SendMessage( ei.hWndEdit, EM_EXSETSEL, 0, (LPARAM) &cr );
                        bNotFound = TRUE;
                    }
                }
                else
                {
                    if ( ((dwSearchFlags & FR_BEGINNING) != FR_BEGINNING) ||
                         (qs_nEditEOF & srchEOF) )
                    {
                        tfW.dwFlags = dwSearchFlags | FR_BEGINNING;
                        if ( !bSearchEx )
                        {
                            iFindResult = (int) SendMessage( g_Plugin.hMainWnd, 
                              AKD_TEXTFIND, (WPARAM) ei.hWndEdit, (LPARAM) &tfW );
                        }
                        else       
                        {
                            iFindResult = doFindTextExW( ei.hWndEdit, &tfW );
                        }
                        if ( iFindResult == -1 )
                        {
                            bNotFound = TRUE;
                        }
                    }
                    else
                    {
                        bNotFound = TRUE;
                    }
                }
                qs_nEditEOF = 0;
            }
        }
        else
        {
            qs_nEditEOF = 0;
        }
    }

    if ( (!bNotFound) && (!qs_nEditEOF) )
    {
        int       nLine;
        int       nLine1st;
        CHARRANGE cr = { 0, 0 };
            
        SendMessage( ei.hWndEdit, EM_EXGETSEL, 0, (LPARAM) &cr );
        nLine = (int) SendMessage(ei.hWndEdit, EM_EXLINEFROMCHAR, 0, cr.cpMin);
        nLine1st = (int) SendMessage(ei.hWndEdit, EM_GETFIRSTVISIBLELINE, 0, 0);
        if ( nLine < nLine1st )
        {
            SendMessage(ei.hWndEdit, EM_LINESCROLL, 0, nLine - nLine1st);
        }
    }
    
    // now the Edit control can be deactivated
    qs_bEditCanBeNonActive = TRUE;
    qsearchDoSetNotFound(hEdit, bNotFound, bEOF);
}

static void CallHighlightMain(void* phlParams)
{
    if ( g_Plugin.bOldWindows )
    {
        PLUGINCALLSENDA pcsA;

        pcsA.pFunction = (char *) cszHighlightMainA;
        pcsA.bOnStart = FALSE;
        pcsA.lParam = (LPARAM) phlParams;
        pcsA.lpbAutoLoad = NULL;

        SendMessageA( g_Plugin.hMainWnd, AKD_DLLCALL, 0, (LPARAM) &pcsA );
    }
    else
    {
        PLUGINCALLSENDW pcsW;

        pcsW.pFunction = (wchar_t *) cszHighlightMainW;
        pcsW.bOnStart = FALSE;
        pcsW.lParam = (LPARAM) phlParams;
        pcsW.lpbAutoLoad = NULL;

        SendMessageW( g_Plugin.hMainWnd, AKD_DLLCALL, 0, (LPARAM) &pcsW );
    }
}

void qsearchDoTryHighlightAll(HWND hDlg)
{
    if ( g_Plugin.bOldWindows )
    {
        if ( 0 == ((LPCSTR) g_QSearchDlg.szFindTextW)[0] )
        {
            // no text to search for: removing the highlight
            qsearchDoTryUnhighlightAll();
            return;
        }
    }
    else
    {
        if ( 0 == ((LPCWSTR) g_QSearchDlg.szFindTextW)[0] )
        {
            // no text to search for: removing the highlight
            qsearchDoTryUnhighlightAll();
            return;
        }
    }
    
    if ( g_bHighlightPlugin && !qs_bEditNotFound )
    {
        HWND hCh = GetDlgItem(hDlg, IDC_CH_HIGHLIGHTALL);

        if ( SendMessage(hCh, BM_GETCHECK, 0, 0) == BST_CHECKED )
        {
            EDITINFO  ei;

            ei.hWndEdit = NULL;
            SendMessage( g_Plugin.hMainWnd, 
            AKD_GETEDITINFO, (WPARAM) NULL, (LPARAM) &ei );
        
            if ( ei.hWndEdit )
            {
                CHARRANGE cr = { 0, 0 };

                SendMessage( ei.hWndEdit, EM_EXGETSEL, 0, (LPARAM) &cr );
                if ( cr.cpMin != cr.cpMax )
                {
                    DLLECHIGHLIGHT_MARK hlParams;
                    BOOL    bMatchCase;
                    wchar_t szTextColor[16];
                    wchar_t szBkgndColor[16];
                    
                    if ( g_Plugin.bOldWindows )
                    {
                        wsprintfA( (char *) szTextColor, "0" );
                        wsprintfA( (char *) szBkgndColor, "#%02X%02X%02X", 
                          GetRValue(g_Options.colorHighlight),
                          GetGValue(g_Options.colorHighlight),
                          GetBValue(g_Options.colorHighlight)
                        );
                    }
                    else
                    {
                        wsprintfW( szTextColor, L"0" );
                        wsprintfW( szBkgndColor, L"#%02X%02X%02X", 
                          GetRValue(g_Options.colorHighlight),
                          GetGValue(g_Options.colorHighlight),
                          GetBValue(g_Options.colorHighlight)
                        );
                    }

                    hCh = GetDlgItem(hDlg, IDC_CH_MATCHCASE);
                    bMatchCase = (SendMessage(hCh, BM_GETCHECK, 0, 0) == BST_CHECKED);

                    hlParams.dwStructSize = sizeof(DLLECHIGHLIGHT_MARK);
                    hlParams.nAction = DLLA_HIGHLIGHT_MARKSELECTION;
                    hlParams.pColorText = (unsigned char *) szTextColor;
                    hlParams.pColorBk = (unsigned char *) szBkgndColor;
                    hlParams.bMatchCase = bMatchCase;
                    hlParams.dwFontStyle = 0;
                    hlParams.dwMarkID = g_Options.dwHighlightMarkID;

                    CallHighlightMain( &hlParams );
                }
                else
                    qsearchDoTryUnhighlightAll(); // no selection
            }
        }
    }
    else
        qsearchDoTryUnhighlightAll(); // not found
}

void qsearchDoTryUnhighlightAll()
{
    if ( g_bHighlightPlugin )
    {
        DLLECHIGHLIGHT_UNMARK hlParams;

        hlParams.dwStructSize = sizeof(DLLECHIGHLIGHT_UNMARK);
        hlParams.nAction = DLLA_HIGHLIGHT_UNMARKSELECTION;
        hlParams.dwMarkID = g_Options.dwHighlightMarkID;

        CallHighlightMain( &hlParams );
    }
}
